#!/usr/bin/env python3

# This script will generate interfacing headers from the official headers of solvers.
# Note that since this is not an actual parser for the headers this is not a foolproof
# method. Please make sure that the generated file do not contain multiple definitions
# of the same identifier. This generally suggest a ifdef statement was read ignored.
#
# Generally this script goes over the header files given as "source_names" and looks
# for declarations of macros, enumerated types, and typedefinitions. It then compares
# each of these definitions to see if they are used in the files declared in "usage".
# Additional control is added by the "ignore" and "include" fields to correct
# the generated header file.

import os
import re
import sys
from itertools import chain
from pathlib import Path
from typing import List, Optional

STDCALL_PREPEND = """#if !defined(WIN32) && !defined(WIN64)
#define __cdecl
#define __stdcall
#endif
"""


def find_file(file_name: str, search_path: List[Path]) -> Optional[Path]:
    for p in search_path:
        path = p / file_name
        if path.exists():
            return path
    return None


def remove_comments(code: str):
    code = re.sub(r"/\*(.|\n)*?\*/", "", code)
    code = re.sub(r"//.*", "", code)
    return code


# Generator for the definitions in the header of the given Path.
def definitions(path: Path):

    with path.open() as file:
        end = True
        buffer = ""
        for line in file.readlines():
            buffer += line

            if line.endswith("\\\n"):
                buffer = buffer[:-2]
                continue

            # macro definitions
            match = re.search(r"#define\s+(\w+).*[ \t]+\S+[^\n]*\n", buffer)
            if match:
                text = remove_comments(match[0])
                yield (match[1], text)
                buffer = ""
                continue
            buffer = ""

    txt = path.read_text()
    # macro definitions
    for match in re.findall(r"^\s*(enum\s+(\w+)([^;]|\n)*;)", txt, flags=re.MULTILINE):
        text = remove_comments(match[0]) + "\n"
        yield (match[1], text)
    # type definitions
    for match in re.findall(r"(typedef\s+.+\s+\*?\s*(\w+)\s*;)", txt):
        text = remove_comments(match[0]) + "\n"
        yield (match[1], text)


if __name__ == "__main__":
    if len(sys.argv) <= 1:
        print(
            "Header generator usage:\n\tgen_header.py <solver> [search_location...]",
            file=sys.stderr,
        )
        exit(1)
    solver = sys.argv[1].upper()
    root = Path(os.path.realpath(__file__)).parent.parent.parent

    # Set addition search paths provided from command line
    search_path = []
    for i in range(2, len(sys.argv)):
        search_path.append(Path(sys.argv[i]).resolve())
        search_path.append(Path(sys.argv[i]).resolve() / "include")

    # Set all solver specific settings
    output = root / "include/minizinc/_thirdparty"
    prepend = ""
    ignore = set()  # identifiers to be ignore (e.g., platform dependent definitions)
    include = (
        set()
    )  # identifiers that must be included (e.g., identifiers used by included macros)

    if solver == "CPLEX":
        output /= "cplex_interface.h"
        source_names = ["ilcplex/cpxconst.h"]
        prepend = """#ifdef _MSC_VER
   typedef __int64 CPXLONG;
#else
   typedef long long CPXLONG;
#endif
#ifdef _WIN32
#define CPXPUBLIC      __stdcall
#else
#define CPXPUBLIC
#endif
"""
        ignore = {"CPXLONG", "CPXPUBLIC"}
        usage = [
            root / "solvers/MIP/MIP_cplex_solverfactory.cpp",
            root / "solvers/MIP/MIP_cplex_wrap.cpp",
            root / "include/minizinc/solvers/MIP/MIP_cplex_solverfactory.hh",
            root / "include/minizinc/solvers/MIP/MIP_cplex_wrap.hh",
        ]

        for ver in [
            "1210",
            "129",
            "128",
            "1271",
            "127",
            "1263",
            "1262",
            "1261",
            "126",
            "201",
        ]:
            paths = [
                f"/opt/ibm/ILOG/CPLEX_Studio{ver}",
                f"/opt/IBM/ILOG/CPLEX_Studio{ver}",
                f"C:\\Program Files\\IBM\\ILOG\\CPLEX_Studio{ver}",
                f"C:\\Program Files (x86)\\IBM\\ILOG\\CPLEX_Studio{ver}",
                f"~/Applications/IBM/ILOG/CPLEX_Studio{ver}",
                f"/Applications/IBM/ILOG/CPLEX_Studio{ver}",
                f"/Applications/CPLEX_Studio{ver}",
            ]
            search_path.extend([Path(p) / "include" for p in paths if Path(p).exists()])
            search_path.extend(
                [Path(p) / "cplex/include" for p in paths if Path(p).exists()]
            )

    elif solver == "GUROBI":
        output /= "gurobi_interface.h"
        source_names = ["gurobi_c.h"]
        prepend = STDCALL_PREPEND
        usage = [
            root / "solvers/MIP/MIP_gurobi_solverfactory.cpp",
            root / "solvers/MIP/MIP_gurobi_wrap.cpp",
            root / "include/minizinc/solvers/MIP/MIP_gurobi_solverfactory.hh",
            root / "include/minizinc/solvers/MIP/MIP_gurobi_wrap.hh",
        ]
        for ver in [
            "913",
            "912",
            "911",
            "910",
            "903",
            "902",
            "901",
            "900",
            "811",
            "810",
            "801",
            "752",
            "702",
        ]:
            paths = [
                f"/opt/gurobi{ver}/linux64",
                f"C:\\gurobi{ver}\\win64",
                f"C:\\gurobi{ver}\\win32",
                f"/Library/gurobi{ver}/mac64",
                f"/Library/gurobi{ver}/macos_universal2",
            ]
            search_path.extend([Path(p) / "include" for p in paths if Path(p).exists()])
    elif solver == "SCIP":
        output /= "scip_interface.h"
        source_names = [
            "scip/def.h",
            "scip/type_scip.h",
            "scip/type_retcode.h",
            "scip/type_message.h",
            "scip/type_prob.h",
            "scip/type_var.h",
            "scip/type_stat.h",
            "scip/type_sol.h",
            "scip/type_event.h",
            "scip/type_cons.h",
            "scip/type_paramset.h",
            "scip/type_lp.h",
            "scip/cons_orbitope.h",
            "symmetry/type_symmetry.h"
        ]
        include = {
            "SCIP_Retcode",
            "SCIP_Vartype",
            "SCIP_Status",
            "SCIP_Objsense",
            "SCIP_BoundType",
            "SCIP_OrbitopeType",
        }
        prepend = """#if !defined(_MSC_VER) || _MSC_VER > 1600
#ifdef __cplusplus
#define __STDC_FORMAT_MACROS
#endif
#include <inttypes.h>
#else
#define PRIx64 "llx"
#endif
#ifdef SCIP_DEBUG
#define SCIPdebugMessage                printf("[%s:%d] debug: ", __FILE__, __LINE__), printf
#else
#define SCIPdebugMessage                while( FALSE ) printf
#endif
"""
        usage = [
            root / "solvers/MIP/MIP_scip_solverfactory.cpp",
            root / "solvers/MIP/MIP_scip_wrap.cpp",
            root / "include/minizinc/solvers/MIP/MIP_scip_solverfactory.hh",
            root / "include/minizinc/solvers/MIP/MIP_scip_wrap.hh",
        ]
    elif solver == "XPRESS":
        output /= "xpress_interface.h"
        source_names = ["xprb.h", "xprs.h"]
        prepend = """#ifdef _WIN32
#define XPRS_CC __stdcall
#else
#define XPRS_CC
#endif
#if defined(_WIN32) || defined(_WIN64)
#define XB_CC __stdcall
#else
#define XB_CC
#endif
#ifdef _WIN32
#define XPRSint64 __int64
#elif defined(__LP64__) || defined(_LP64) || defined(__ILP64__) || defined(_ILP64)
#define XPRSint64 long
#else
#define XPRSint64 long long
#endif
"""
        ignore = {"XB_CC", "XPRS_CC", "XPRSint64"}
        include = {
            "XB_INFINITY",
            "XB_PL",
            "XB_BV",
            "XB_UI",
            "XB_L",
            "XB_G",
            "XB_E",
            "XB_MAXIM",
            "XB_MINIM",
            "XB_LP",
            "XB_MPS",
            "XB_XO_SOL",
            "XB_MIP_NOT_LOADED",
            "XB_MIP_NO_SOL_FOUND",
            "XB_MIP_INFEAS",
            "XB_MIP_OPTIMAL",
            "XB_MIP_UNBOUNDED",
        }
        usage = [
            root / "solvers/MIP/MIP_xpress_solverfactory.cpp",
            root / "solvers/MIP/MIP_xpress_wrap.cpp",
            root / "include/minizinc/solvers/MIP/MIP_xpress_solverfactory.hh",
            root / "include/minizinc/solvers/MIP/MIP_xpress_wrap.hh",
        ]
        search_path.extend(
            [
                "/opt/xpressmp/include",
                "C:\\xpressmp\\Applications\\FICO Xpress\\xpressmp\\include",
            ]
        )
    else:
        print(
            f"ERROR: solver `{sys.argv[1]}' unknown. Unable to generate headers",
            file=sys.stderr,
        )
        exit(1)

    # Find Source Files
    source_files = []
    for name in source_names:
        f = find_file(name, search_path)
        if f is None:
            print(
                f"ERROR: Unable to locate source file `{name}'. Did you forget to specify a search location?",
                file=sys.stderr,
            )
            exit(1)
        source_files.append(f)

    # Filter definitions on usage and place in generated header file
    combined_usage = "\n".join([f.read_text() for f in usage])
    with output.open("w") as out:
        out.write(prepend)
        for (ident, define) in chain(*[definitions(f) for f in source_files]):
            if ident in include or (ident not in ignore and ident in combined_usage):
                out.write(define)



from . import yaml


@yaml.sequence(u"!Unordered")
class Unordered:
    """
    A list where order is not important (like a set, except elements can be repeated)

    Represented by `!Unordered` in YAML.
    """

    def __init__(self, *args):
        self.items = args

    def __iter__(self):
        return iter(self.items)

    def __eq__(self, other):
        items = list(self.items)
        try:
            for x in other:
                items.remove(x)
            return len(items) == 0
        except ValueError:
            return False


@yaml.scalar(u"!Approx")
class Approx:
    """
    A helper which allows for approximate comparison of floats using `!Approx`
    """

    def __init__(self, value, threshold=0.000001):
        self.value = float(value)
        self.threshold = threshold

    def __eq__(self, other):
        try:
            diff = abs(float(other) - self.value)
            return diff <= self.threshold
        except (TypeError, ValueError):
            return False

    def __repr__(self):
        return "Approx({})".format(self.value)

    def get_value(self):
        return self.value


@yaml.scalar(u"!Trim")
class Trim:
    """
    A helper which allows for comparison of trimmed strings with `!Trim`
    """

    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        return self.value.strip() == other.strip()

    def __repr__(self):
        return "Strip({})".format(self.value)

    def get_value(self):
        return self.value



##  This class compares and summarizes instance solution results
##  for a given set of optimization/satisfiability instances.
##  Looks for contradictions. Heuristic performance indicators. 
##  (c) Gleb Belov@monash.edu  2017

from collections import OrderedDict
# import prettytable
import utils, io  ## functools
from utils import strNL

#####################################################################################
################### class CompareLogs ##################
#####################################################################################

## It receives a list of dictionaries, keyed by instances files,
## containing values of solutions status etc., and performs the checks + summary.
class CompareLogs:
    def __init__( self ):
        self.lResLogs = []                                 ## empty list of logs/methods to compare
        self.hdrSummary = [       ## These are column headers for overall per-method summary
            ( "s_MethodName", "logfile/test name", "Logfile and possibly test alias" ),
            ( "n_Reported",   "Nout", "Total number of instances" ),
            ( "n_CheckFailed","Nbad", "Number of failed solution checks" ),
            ( "n_ErrorsBackend",     "NerrB", "Number of backend errors. TODO need still to consider feasible solutions if ERROR status" ),
            ( "n_ErrorsLogical",     "NerrL", "Number of logical errors, such as different solver and MZN obj values" ),
            ( "n_OPT",        "Nopt", "Number of reported optimal" ),
            ( "n_FEAS",       "Nfea", "Number of reported feasible" ),
            ( "n_SATALL",     "NsatA", "Number of reported SAT-COMPLETE" ),
            ( "n_SAT",        "Nsat", "Number of reported SAT" ),
            ( "n_INFEAS",     "Ninfeas", "Number of reported UNSAT" ),
            ( "n_NOFZN",      "NoFZN", "Number of failed flattenings" ),
            ( "n_UNKNOWN",    "Nunkn", "Number of unknown results" ),
            ( "t_Flatten",    "TFlt", "Total flattening time" )
          ]
        self.hdrRanking = [       ## These are column headers for ranking analysis
            ## ( "nmMeth", "logfile/test/method name" ),
            ( "ONFZ",   "Number of instances where ONLY this method failed to compile (NOFZN)" ),
            ( "OOpt",   "Number of instances where ONLY this method is OPTIMAL" ),
            ( "OSaC",   "Number of instances where ONLY this method is SAT-COMPLETE" ),
            ( "OFeas",  "Number of instances where ONLY this method is FEASIBLE and none is optimal" ),
            ( "OSat",   "Number of instances where ONLY this method is SAT and none is optimal" ),
            ( "OInfeas","Number of instances where ONLY this method is INFeasible" ),
            ( "BPri",   "Number of instances where this method has a better PRIMAL BOUND" ),
            ( "BDua",   "Number of instances where this method has a better DUAL BOUND" )
            ##  TODO: ranks here ( "n_UNKNOWN",    "Nunkn" )
          ]
 
    hdrTable = OrderedDict( [                  ## Possible headers for table printout
            ( "stt", "The solver status" ),  ## TODO an error should be separate flag, not a status
            ( "chk", "The solution checking status" ),
            ( "objMZN", "The MZN obj value" ),
            ( "objSLV", "The solver obj value" ),
            ( "bnd", "The solver dual bound" ),
            ( "tAll", "Total running wall time" ),
            ( "tFlt", "Flattening time" ),
            ( "tBest", "A best solution's finding time" ),
            ( "sns", "Model sense (min/max/sat)" ),
            ( "errH", "Solver errors" ),
            ( "errL", "Logical errors" )
            ] )

    ## which of those to print for each method 
    hdrTable2P = "stt chk objMZN objSLV bnd tFlt tBest"
    hdrTable2P_spl = hdrTable2P.split( " " )

    mapStatShort = {              ## Short status names
            2:  "OPT",
            1:  "FEAS",
            4:  "SATA",
            3:  "SAT",
            0:  "UNKN",
            -1: "UNSAT",
            -2: "UNBND",
            -3: "UNSorUNB",
            -4: "ERR_H"
            }

    mapProblemSense = {
            1:  "max",
            0:  "sat",
            -1: "min",
            -2: "nosns"
            }

    ## Add a method's log
    def addLog( self, sName ):
        self.lResLogs.append( ( OrderedDict(), [ sName, '' ] ) )      ## a tuple of dict and list of names ([filename, nick])
        return self.getLastLog()
        
    def getLastLog( self ):
        assert 0<len( self.lResLogs )
        return self.lResLogs[-1]

    def getMethodName( self, lNames ):    ## produce a sigle string from a list to create a method name
        return " ".join( lNames )
      
    ## Return the union of all instances in all logs
    def getInstanceUnion( self ):
    ##    return list( functools.reduce(set.union, (set(d[0].keys()) for d in self.lResLogs)) )
        r0 = OrderedDict()
        for d in self.lResLogs:
            r0.update( OrderedDict.fromkeys( d[0] ) )
        return r0.keys()

    ## Return the intersection of all instances in all logs
    def getInstanceIntersection( self ):
        assert False
        return []                      ## TODO

    ## This compares all instances specified in the list of pairs (full name, short name)
    ## If comparing by-instance from outside, follow this pattern
    def compareAllInstances( self, lInstances ):
        self.initListComparison()
        for sInst in lInstances:
            self.compareInstance( sInst )
        self.summarizeCmp()
        self.summarizeFinalHdr()
        self.summarize()
        
    ## Init stats etc.
    def initListComparison( self ):
        self.nInstCompared = 0
        self.nInstWithOptSense = 0
        ## Using warning strings...     self.lInstContradOptSense = []
        self.lCmpVecs = []                  # List of summary vectors for each method
        self.mCmpVecVals = {}               # Maps to the "quantity" parts of those
        self.mCmpVecQual = {}               # Maps to the "quality" parts
        print( "" )                                        ## Newline
        print( "=============== PER-INSTANCE RESULTS TABLE, HEADERS: ===============" )
        for hdrLine in self.hdrTable.items():
            print( "     ", hdrLine )
        print( "====================================================================" )
        print( "No.\tinst", end='\t')
        for mLog, lN in self.lResLogs:                ## Select method and its name list
            lNames = self.getMethodName(lN)
            av = OrderedDict({ "s_MethodName": lNames })
            aq = OrderedDict({ "s_MethodName": lNames })
            self.lCmpVecs.append( ( lNames, av, aq ) )
            self.mCmpVecVals[ lNames ] = av
            self.mCmpVecQual[ lNames ] = aq
            for hdr in self.hdrTable2P_spl:                    ## Print table headers
                print( hdr, end='\t' )

        print( "" )                                        ## Newline
            
        self.mInfeas, self.mNoFZN, self.mFail, self.mError = {}, {}, {}, {}
        self.nContrStatus, self.nContrOptVal, self.nContrBounds = 0, 0, 0
        self.matrRanking = utils.MatrixByStr(
                [ ( self.getMethodName( lNames ), "No long name" ) for mLog, lNames in self.lResLogs ],
                self.hdrRanking )
        self.matrRankingMsg = utils.MatrixByStr(
                [ ( self.getMethodName( lNames ), "No long name" ) for mLog, lNames in self.lResLogs ],
                self.hdrRanking, [] )

        self.nNoOptAndAtLeast2Feas = 0

        ############################## Output strings. TODO into a map
        self.ioContrSense = io.StringIO() 
        self.ioContrStatus = io.StringIO()
        self.ioContrObjValMZN = io.StringIO()
        self.ioBadObjValueStatusOpt = io.StringIO()
        self.ioBadObjValueStatusFeas = io.StringIO()
        self.ioContrOptVal = io.StringIO()
        self.ioContrBounds = io.StringIO()
        self.ioBadChecks = io.StringIO()
        self.ioErrors = io.StringIO()
      
    ## Compare methods on the given instance
    def compareInstance( self, sInst ):
        self.initInstanceComparison( sInst[0] )
        self.tryFindProblemSense( sInst[0] )
        self.compileInstanceResults( sInst[0] )
        self.checkContradictions( sInst[0] )
        self.rankPerformance( sInst[0] )
        self.doInstanceSummary( sInst )
      
    ## Summarize up to current instance
    def summarizeCurrent( self, lLogNames ):
        lNames = self.getMethodName(lLogNames)
        mCmpVals = self.mCmpVecVals[lNames]
        for hdr in self.hdrSummary:
            print( hdr[1], ":",
                  mCmpVals[hdr[0]] if hdr[0] in mCmpVals else 0,
                  sep='', end=' ' )

    ## Summarize comparisons / ranking
    def summarizeCmp( self ):
        print(
            self.matrRankingMsg.stringifyLists( "       METHODS' STAND-OUTS",
                    "        METHOD",
                    "        PARAMETER" ) + \
            "\n------------------ SOLUTION CHECKS FAILURES ------------------\n\n" + \
            self.ioBadChecks.getvalue() + \
            "\n\n------------------ OBJECTIVE SENSE CONTRADICTIONS ------------------\n\n" + \
            self.ioContrSense.getvalue() + \
            "\n\n------------------ OBJECTIVE VALUE MZN / SOLVER CONTRADICTIONS ------------------\n\n" + \
            self.ioContrObjValMZN.getvalue() + \
            "\n\n------------------ OBJECTIVE VALUE BAD, STATUS OPTIMAL ------------------\n\n" + \
            self.ioBadObjValueStatusOpt.getvalue() + \
            "\n\n------------------ OBJECTIVE VALUE BAD, STATUS FEASIBLE ------------------\n\n" + \
            self.ioBadObjValueStatusFeas.getvalue() + \
            "\n\n------------------ STATUS CONTRADICTIONS ------------------\n\n" + \
            self.ioContrStatus.getvalue() + \
            "\n\n------------------ OBJECTIVE VALUE CONTRADICTIONS ------------------\n\n" + \
            self.ioContrOptVal.getvalue() + \
            "\n\n------------------ DUAL BOUND CONTRADICTIONS ------------------\n\n" + \
            self.ioContrBounds.getvalue() + \
            "\n\n------------------ ERRORS REPORTED BY SOLVERS ------------------\n\n" + \
            self.ioErrors.getvalue() + "\n" + \
            "\n\n------------------ RANKING ------------------\n\n" + \
            "\n".join( [ "     " + str( hl ) for hl in self.hdrRanking ] ) + \
            "\n---------------------------------------------\n" + \
            self.matrRanking.stringify2D()
        )

    ## Summary headers
    def summarizeFinalHdr( self ):
        return \
            "\n".join( [ "     " + str((hdrLine[1], hdrLine[2])) for hdrLine in self.hdrSummary ] ) + \
            "\n=================================================="
    
    ## Summarize
    def summarize( self ):
        return \
            utils.MyTab().tabulate(
              [ [ lcv[1][hdr[0]] if hdr[0] in lcv[1] else 0
                  for hdr in self.hdrSummary ]
                for lcv in self.lCmpVecs ],
                [ pr[1] for pr in self.hdrSummary ]
            )
      
###############################################################################################
####################### LEVEL 2 #########################
###############################################################################################
    def initInstanceComparison( self, sInst ):
        self.lNOFZN, self.lOpt, self.lSatAll, self.lFeas, self.lSat, self.lInfeas = [], [], [], [], [], []
        self.mOptVal, self.lOptVal, self.lPrimBnd, self.lDualBnd = OrderedDict(), [], [], []
        self.nInstCompared += 1
        self.nReported = 0               ## How many methods reported for this instances
        ## Detailed table line for this instance 
        self.aDetThisInst = { self.getMethodName( ll[1] ) : {} for ll in self.lResLogs  }

    def tryFindProblemSense( self, sInst ):
        self.sSenses = {}
        self.nOptSenseGiven = -2;
        for mLog, lNames in self.lResLogs:
            if sInst in mLog:
                mSlv = mLog[ sInst ][ "__SOLVE__" ]               ## __SOLVE__ always there???
                if "Problem_Sense" in mSlv:
                    self.sSenses[ mSlv["Problem_Sense"][0] ] = lNames # mSlv["Problem_Sense"][1]
        if 1<len( self.sSenses ):
            print( "WARNING: DIFFERENT OBJ SENSES REPORTED for the instance ", sInst,
                   ":  ", self.sSenses, sep='', file=self.ioContrSense )
        elif 1==len( self.sSenses ):
            self.nOptSenseGiven = list(self.sSenses.keys())[0]
                
    def compileInstanceResults( self, sInst ):
        for mLog, lN in self.lResLogs:                ## Select method and its name list
            lNames = self.getMethodName(lN)
            aDetThis = self.aDetThisInst[ lNames ]    ## Result table line section
            if sInst in mLog:
                self.nReported += 1
                aResultThisInst = OrderedDict({ "n_Reported": 1 })
                aResultThisInst[ "n_CheckFailed" ] = 0
                mRes = mLog[ sInst ]                      ## The method's entry for this instance
                aDetThis[ "chk" ] = "ok"
                if "SOLUTION_CHECKS_FAILED" in mRes and \
                  0<mRes["SOLUTION_CHECKS_FAILED"]:
                    aResultThisInst[ "n_CheckFailed" ] = 1
                    aDetThis[ "chk" ] = "BAD"
                    print( "WARNING: SOLUTION CHECK(S) FAILED for the instance ", sInst,
                           ",  method '", lNames, "'.", sep='', file = self.ioBadChecks )
                    ## utils.addMapValues( self.mCmpVecVals[lNames], aResultThisInst )
                    ## continue                                        ## TODO. Param?
                aResultThisInst[ "n_ErrorsBackend" ] = 0
                aResultThisInst[ "n_ErrorsLogical" ] = 0
                aDetThis [ "errH" ] = 0
                aDetThis [ "errL" ] = 0
                mSlv = mRes[ "__SOLVE__" ]
                dObj_MZN = utils.try_float( mSlv.get( "ObjVal_MZN" ) )
                aDetThis [ "objMZN" ] = dObj_MZN
                dObj_SLV = utils.try_float( mSlv.get( "ObjVal_Solver" ) )
                aDetThis [ "objSLV" ] = dObj_SLV
                dBnd_SLV = utils.try_float( mSlv.get( "DualBnd_Solver" ) )
                aDetThis [ "bnd" ] = dBnd_SLV
                dTime_All = utils.try_float( mSlv.get( "TimeReal_All" ) )
                aDetThis [ "tAll" ] = dTime_All
                dTime_Flt = utils.try_float( mSlv.get( "Time_Flt" ) )
                aResultThisInst[ "t_Flatten" ] = dTime_Flt if dTime_Flt is not None else dTime_All  ##?? Assume flattening stopped
                aDetThis [ "tFlt" ] = dTime_Flt
                dTime_Last = utils.try_float( mSlv.get( "TimeReal_LastStatus" ) )
                aDetThis [ "tBest" ] = dTime_Last
                ## Compare obj vals
                dObj, bObj_MZN = (dObj_MZN, True) if \
                      None!=dObj_MZN and abs( dObj_MZN ) < 1e45 else (dObj_MZN, False)
                ## Assuming solver value is better if different. WHY? Well it' happened both ways
                dObj, bObj_SLV = (dObj_SLV, True) if \
                      None!=dObj_SLV and abs( dObj_SLV ) < 1e45 else (dObj, False)
                if bObj_MZN and bObj_SLV:
                    if abs( dObj_MZN-dObj_SLV ) > 1e-6 * max( abs(dObj_MZN), abs(dObj_SLV) ):
                        aResultThisInst[ "n_ErrorsLogical" ] += 1
                        aDetThis [ "errL" ] += 1
                        print ( "  WARNING: DIFFERENT MZN / SOLVER OBJ VALUES for the instance ", sInst,
                           ", method '", lNames, "' : ",
                           dObj_MZN, " / ", dObj_SLV, sep='', file=self.ioContrObjValMZN)
                ## Retrieve solution status
                if "Sol_Status" in mSlv:
                    n_SolStatus = mSlv[ "Sol_Status" ][0]
                else:
                    n_SolStatus = 0
                ## Retrieve dual bound
                dBnd = None
                if None!=dBnd_SLV and abs( dBnd_SLV ) < 1e45:
                    dBnd = dBnd_SLV
                    self.lDualBnd.append( ( dBnd_SLV, lNames ) )      ## Even infeas instances can have dual bound?
                ## Trying to deduce opt sense if not given:
                if 1==len(self.sSenses):
                    nSense = next(iter(self.sSenses.keys()))
                else:
                    nSense = -2  ## ??
                aDetThis[ "sns" ] = self.mapProblemSense[ nSense ]
                self.bOptProblem = True if 0!=nSense else False     ## or (None!=dBnd or None!=dObj)
                    ### ... here assumed it's an opt problem by default... why... need to check bounds first??
                ## Handle optimality / SAT completed
                if 2==n_SolStatus:
                    if not self.bOptProblem:
                        self.lSatAll.append( lNames )
                        aResultThisInst[ "n_SATALL" ] = 1
                        aDetThis[ "stt" ] = self.mapStatShort[ 4 ]
                    else:                                        ## Assume it's an optimization problem????? TODO
                        self.lOpt.append( lNames )                   ## Append the optimal method list
                        aResultThisInst[ "n_OPT" ] = 1
                        aDetThis[ "stt" ] = self.mapStatShort[ 2 ]
                        if dObj is None or abs( dObj ) >= 1e45:
                            aResultThisInst[ "n_ErrorsLogical" ] += 1
                            aDetThis [ "errL" ] += 1
                            print ( "  WARNING: OPTIMAL STATUS BUT BAD OBJ VALUE, instance ", sInst,
                              ", method '", lNames, "': '",
                              ( "" if None==dObj else str(dObj) ), "', result record: ",   # mRes,
                              ",, dObj_MZN: ", dObj_MZN, sep='', file=self.ioBadObjValueStatusOpt )
                        else:
                            self.mOptVal[ dObj ] = lNames            ## Could have used OrderedDict
                            self.lOptVal.append( (dObj, lNames) )    ## To have both a map and the order
                            self.lPrimBnd.append( (dObj, lNames) )
                ## Handle feasibility / SAT
                elif 1==n_SolStatus:
                    if not self.bOptProblem:
                        self.lSat.append( lNames )
                        aResultThisInst[ "n_SAT" ] = 1
                        aDetThis[ "stt" ] = self.mapStatShort[ 3 ]
                    else:                                        ## Assume it's an optimization problem????? TODO
                        self.lFeas.append( lNames )                   ## Append the optimal method list
                        aResultThisInst[ "n_FEAS" ] = 1
                        aDetThis[ "stt" ] = self.mapStatShort[ 1 ]
                        if None==dObj or abs( dObj ) >= 1e45:
                            aResultThisInst[ "n_ErrorsLogical" ] += 1
                            aDetThis [ "errL" ] += 1
                            print ( "  WARNING: feasible status but bad obj value, instance ", sInst,
                                    ", method '", lNames, "' :'",
                              ( "" if None==dObj else str(dObj) ), "', result record: ",  #  mRes,
                              sep='', file=self.ioBadObjValueStatusFeas )
                        else:
                            self.lPrimBnd.append( (dObj, lNames) )
                ## Handle infeasibility
                elif -1>=n_SolStatus and -3<=n_SolStatus:
                    self.lInfeas.append( lNames )
                    aResultThisInst[ "n_INFEAS" ] = 1
                    aDetThis[ "stt" ] = self.mapStatShort[ n_SolStatus ]
                    self.mInfeas. setdefault( sInst, [] )
                    self.mInfeas[ sInst ].append( lNames )
                ## Handle ERROR?
                elif -4==n_SolStatus:
                    aResultThisInst[ "n_ErrorsBackend" ] = 1
                    aDetThis [ "errH" ] += 1
                    aDetThis[ "stt" ] = self.mapStatShort[ n_SolStatus ]   ## Should not happen TODO
                    self.mError. setdefault( sInst, [] ).append( lNames )
                    print( "ERROR REPORTED for the instance ", sInst, ", method '", lNames,
                            "',  result record: ",   ## mRes,
                            sep='', file=self.ioErrors )
                else:
                    aResultThisInst[ "n_UNKNOWN" ] = 1
                    aDetThis[ "stt" ] = self.mapStatShort[ 0 ]
                ## Handle NOFZN
                if None==dTime_Flt:
                    aResultThisInst[ "n_NOFZN" ] = 1
                    self.mNoFZN. setdefault( sInst, [] ).append( lNames )
                    self.lNOFZN.append( lNames )
                ## Handle FAIL???
                # LAST:
                utils.addMapValues( self.mCmpVecVals[lNames], aResultThisInst )


    ###
    ### Now compare between differen methods: CONTRADICTIONS
    ###
    def checkContradictions( self, sInst ):
        self.fContr = False
        if len(self.lOpt)+len(self.lFeas)+len(self.lSatAll)+len(self.lSat) > 0 and len(self.lInfeas) > 0:
            self.nContrStatus += 1
            self.fContr = True
            print(  "CONTRADICTION of STATUS: instance " + str(sInst) + ": " + \
                   "\n  OPTIMAL:  " + strNL( "\n       ", self.lOpt) + \
                   "\n  FEAS:  " + strNL( "\n       ", self.lFeas) + \
                   "\n  SAT_COMPLETE:  " + strNL( "\n       ", self.lSatAll) + \
                   "\n  SAT:  " + strNL( "\n       ", self.lSat) + \
                   "\n  INFEAS:  " + strNL( "\n       ", self.lInfeas), file= self.ioContrStatus )
        if len(self.mOptVal) > 1:
            self.nContrOptVal += 1
            self.fContr = True
            print( "CONTRADICTION of OPTIMAL VALUES: " + str(sInst) + \
              ": " + strNL( "\n       ", self.mOptVal.items()), file=self.ioContrOptVal )
        self.nOptSense=0;      ## Take as SAT by default
        if len(self.lPrimBnd)>0 and len(self.lDualBnd)>0 and len(self.lOpt)<self.nReported:
            lKeysP, lValP = zip(*self.lPrimBnd)
            lKeysD, lValD = zip(*self.lDualBnd)
            nPMin, nPMax, nDMin, nDMax = \
                min(lKeysP), max(lKeysP), \
                min(lKeysD), max(lKeysD)
            if nPMax <= nDMin + 1e-6 and nPMin < nDMax - 1e-6:
                self.nOptSense=1                             ## maximize
            elif nPMin >= nDMax - 1e-6 and nPMax > nDMin + 1e-6:
                self.nOptSense=-1                            ## minimize
            elif nPMax > nDMin + 1e-6 and nPMin < nDMax - 1e-6 or \
               nPMin < nDMax - 1e-6 and nPMax > nDMin + 1e-6:
                self.nContrBounds += 1
                self.fContr = True
                print( "CONTRADICTION of BOUNDS: instance " + str(sInst) + \
                  ":\n  PRIMALS: " + strNL( "\n       ", self.lPrimBnd) + \
                  ",\n  DUALS: " + strNL( "\n       ", self.lDualBnd),
                  file = self.ioContrBounds )
            else:
                self.nOptSense=0          ## SAT
            if 1==len(self.sSenses) and self.nOptSense!=0:
                if self.nOptSense!=self.nOptSenseGiven:     ## access the 'given' opt sense
                    print( "CONTRADICITON of IMPLIED OBJ SENSE:  Instance "+ str(sInst) + \
                      ": primal bounds " + strNL( "\n       ", self.lPrimBnd) + \
                      " and dual bounds "+ strNL( "\n       ", self.lDualBnd) + \
                      " together imply opt sense " + str(self.nOptSense) + \
                      ",  while result logs say "+  str(self.nOptSenseGiven), file=self.ioContrBounds )
                ## else accepting nOptSense as it is

      
      
    ###
    ### Now compare between differen methods: DIFFERENCES AND RANKING
    ###
    def rankPerformance( self, sInst ):
        ### Accepting the opt sense from result tables, if given
        if self.nOptSenseGiven!=-2:
            self.nOptSense = self.nOptSenseGiven
        ### Compare methods on this instance:
        if not self.fContr and self.nReported == len(self.lResLogs):
            if len(self.lNOFZN) == 1:
                self.matrRanking[self.lNOFZN[0], "ONFZ"] += 1
                self.matrRankingMsg[self.lNOFZN[0], "ONFZ"].append( \
                  str(sInst) + ":   the ONLY NON-FLATTENED")
            elif len(self.lOpt) == 1:
                self.matrRanking[self.lOpt[0], "OOpt"] += 1
                self.matrRankingMsg[self.lOpt[0], "OOpt"].append( \
                  str(sInst) + ":   the ONLY OPTIMAL")
            elif len(self.lSatAll) == 1:
                self.matrRanking[self.lSatAll[0], "OSaC"] += 1
                self.matrRankingMsg[self.lSatAll[0], "OSaC"].append( \
                  str(sInst) + ":   the ONLY SAT-COMPLETE")
            elif len(self.lOpt) == 0 and len(self.lFeas) == 1:
                self.matrRanking[self.lFeas[0], "OFeas"] += 1
                self.matrRankingMsg[self.lFeas[0], "OFeas"].append( \
                  str(sInst) + ":   the ONLY FEASIBLE")
            elif len(self.lSatAll) == 0 and len(self.lSat) == 1:
                self.matrRanking[self.lSat[0], "OSat"] += 1
                self.matrRankingMsg[self.lSat[0], "OSat"].append( \
                  str(sInst) + ":   the ONLY SAT")
            elif len(self.lOpt) == 0 and len(self.lFeas) > 1:
                self.nNoOptAndAtLeast2Feas += 1
            elif len(self.lInfeas) == 1:
                self.matrRanking[self.lInfeas[0], "OInfeas"] += 1
                self.matrRankingMsg[self.lInfeas[0], "OInfeas"].append( \
                  str(sInst) + ":   the ONLY INFeasible")
        if not self.fContr \
           and 0==len(self.lInfeas) and 1<len(self.lFeas) and 0!=self.nOptSense:
            self.lPrimBnd.sort()
            if self.nOptSense>0:
                self.lPrimBnd.reverse()
            dBnd, dNM = zip(*self.lPrimBnd)
            dBetter = (dBnd[0]-dBnd[1]) * self.nOptSense
            if 1e-2 < dBetter:                   ## Param?  TODO
                self.matrRanking[dNM[0], "BPri"] += 1
                self.matrRankingMsg[dNM[0], "BPri"].append( str(sInst) \
                + ":      the best OBJ VALUE   by " + str(dBetter) \
                + "\n      PRIMAL BOUNDS AVAILABLE: " + strNL( "\n       ", self.lPrimBnd))
        if not self.fContr \
           and 0==len(self.lInfeas) and 1<len(self.lDualBnd) and 0!=self.nOptSense:
            self.lDualBnd.sort()
            if self.nOptSense<0:
                self.lDualBnd.reverse()
            dBnd, dNM = zip(*self.lDualBnd)
            dBetter = (dBnd[1]-dBnd[0]) * self.nOptSense
            if 1e-2 < dBetter:                   ## Param/adaptive?  TODO
                self.matrRanking[dNM[0], "BDua"] += 1
                self.matrRankingMsg[dNM[0], "BDua"].append( str(sInst) \
                + ":      the best DUAL BOUND   by " + str(dBetter) \
                + "\n      DUAL BOUNDS AVAILABLE: " + strNL( "\n       ", self.lDualBnd))
      
    ###
    ### Now print a table line summarizing the instance for different methods
    ###
    def doInstanceSummary( self, sInst ):
        lIL = [ sInst[1] ]                            ## The short instance name
        print( self.nInstCompared, ".\t", sInst[1], sep='', end='\t' )
        for mLog, lN in self.lResLogs:                ## Select method and its name list
            lNames = self.getMethodName(lN)
            aDetThis = self.aDetThisInst[ lNames ]    ## Result table line section
            if sInst[0] in mLog:
                for hdr in self.hdrTable2P_spl:
                    if hdr not in self.hdrTable:
                        print( '?', end='\t' )
                    elif aDetThis.get( hdr ) is None:
                        print( "-", end='\t' )
                    else:
                        print( aDetThis.get(hdr), end='\t' )
            else:
                for hdr in self.hdrTable2P_spl:
                    print( '-' if hdr in self.hdrTable else '?', end='\t' )
        print( "" )                               ## Newline




from minizinc_testing import yaml
from minizinc_testing.spec import CachedResult

import pytest

# pylint: disable=import-error,no-name-in-module
from py.xml import html
from html import escape
import pytest_html
import re
import minizinc as mzn
from difflib import HtmlDiff
import sys


def pytest_configure(config):
    pytest.solver_cache = {}
    search = config.getoption("--driver")
    if search is not None:
        driver = mzn.Driver.find([search])
        if driver is None:
            raise Exception("Failed to find MiniZinc driver in {}".format(search))
        driver.make_default()


def pytest_addoption(parser):
    parser.addoption(
        "--solvers",
        action="store",
        metavar="SOLVERS",
        help="only run tests with the comma separated SOLVERS.",
    )
    parser.addoption(
        "--suite",
        action="append",
        default=[],
        metavar="SUITE_NAME",
        help="Use the given YAML configuration from suites.yml",
    )
    parser.addoption(
        "--all-suites", action="store_true", dest="feature", help="Run all test suites"
    )
    parser.addoption(
        "--driver",
        action="store",
        metavar="MINIZINC",
        help="Directory containing MiniZinc executable",
    )


def pytest_collect_file(parent, file_path):
    if file_path.suffix == ".mzn":
        return MznFile.from_parent(parent, path=file_path)


def pytest_html_results_table_header(cells):
    cells.insert(2, html.th("Solver", class_="sortable", col="solver"))
    cells.insert(3, html.th("Checker", class_="sortable", col="checker"))
    cells.pop()


def pytest_html_results_table_row(report, cells):
    if hasattr(report, "user_properties"):
        props = {k: v for k, v in report.user_properties}
        cells.insert(2, html.td(props.get("solver", "-")))
        cells.insert(3, html.td(props.get("checker", "-")))
    cells.pop()


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_makereport(item, call):
    outcome = yield
    report = outcome.get_result()
    extra = getattr(report, "extra", [])
    if report.when == "call" and report.outcome != "skipped":
        props = {k: v for k, v in report.user_properties}
        if "compare" in props:
            required, obtained = props["compare"]
            html_content = """
                <button class="copy-button" onclick="this.nextElementSibling.select();document.execCommand('copy');this.textContent = 'Copied!';">Copy obtained output to clipboard</button>
                <textarea class="hidden-textarea" readonly>{}</textarea>
            """.format(
                escape(obtained)
            )
            actual = obtained.split("\n")
            htmldiff = HtmlDiff(2)
            html_content += '<h4>Diffs</h4><div class="diffs">'
            html_content += "<hr>".join(
                htmldiff.make_table(
                    expected.split("\n"),
                    actual,
                    fromdesc="expected",
                    todesc="actual",
                    context=True,
                )
                for expected in required
            )
            html_content += "</div>"
            extra.append(pytest_html.extras.html(html_content))
        report.extra = extra


def pytest_metadata(metadata):
    # Ensure that secrets don't get shown
    # Can likely be removed after pytest-metadata is updated
    metadata.pop("CI_JOB_TOKEN", None)
    metadata.pop("CI_REPOSITORY_URL", None)
    metadata.pop("CI_REGISTRY_PASSWORD", None)


class MznFile(pytest.File):
    def collect(self):
        with open("./spec/suites.yml", encoding="utf-8") as suites_file:
            suites = yaml.load(suites_file)

            if not self.config.getoption("--all-suites"):
                enabled_suites = self.config.getoption("--suite")
                if len(enabled_suites) == 0:
                    suites = {"default": suites["default"]}
                else:
                    suites = {k: v for k, v in suites.items() if k in enabled_suites}

        with self.fspath.open(encoding="utf-8") as file:
            contents = file.read()
            yaml_comment = re.match(r"\/\*\*\*\n(.*?)\n\*\*\*\/", contents, flags=re.S)
            if yaml_comment is None:
                pytest.skip(
                    "skipping {} as no tests specified".format(str(self.fspath))
                )
            else:
                tests = [doc for doc in yaml.load_all(yaml_comment.group(1))]

                for suite_name, suite in suites.items():
                    if any(
                        self.fspath.fnmatch(glob) for glob in suite.includes
                    ) and not any(self.fspath.fnmatch(glob) for glob in suite.excludes):
                        for i, spec in enumerate(tests):
                            for solver in spec.solvers:
                                base = (
                                    str(i) if spec.name is yaml.Undefined else spec.name
                                )
                                name = "{}.{}.{}".format(suite_name, base, solver)
                                cache = CachedResult()
                                yield SolveItem.from_parent(
                                    self,
                                    name=name,
                                    spec=spec,
                                    solver=solver,
                                    cache=cache,
                                    markers=spec.markers,
                                    suite=suite,
                                )
                                for checker in spec.check_against:
                                    yield CheckItem.from_parent(
                                        self,
                                        name="{}:{}".format(name, checker),
                                        cache=cache,
                                        solver=solver,
                                        checker=checker,
                                        markers=spec.markers,
                                        suite=suite,
                                    )


class MznItem(pytest.Item):
    def __init__(self, name, parent, solver, markers, suite):
        super().__init__(name, parent)
        self.user_properties.append(("solver", solver))
        for marker in markers:
            self.add_marker(marker)

        if self.config.getoption("--solvers") is not None:
            solvers_option = self.config.getoption("--solvers").split(",")
            if suite.solvers is None:
                self.allowed = solvers_option
            else:
                self.allowed = list(
                    set(suite.solvers).intersection(set(solvers_option))
                )
        else:
            self.allowed = suite.solvers

        if not self.solver_allowed(solver):
            self.add_marker(
                pytest.mark.skip("skipping {} not in {}".format(solver, self.allowed))
            )

        if not self.solver_exists(solver):
            self.add_marker(pytest.mark.skip("Solver {} not available".format(solver)))

    def solver_allowed(self, solver):
        return self.allowed is None or solver in self.allowed

    def solver_exists(self, solver):
        if solver.endswith(".msc"):
            return True
        solver_exists = pytest.solver_cache.get(solver, None)
        if solver_exists is None:
            try:
                s = mzn.Solver.lookup(solver)
                empty_model = mzn.Model()
                empty_model.add_string("solve satisfy;")
                instance = mzn.Instance(s, empty_model)
                instance.solve()
                solver_exists = True
            except (mzn.MiniZincError, LookupError) as error:
                solver_exists = False
            finally:
                pytest.solver_cache[solver] = solver_exists
        return solver_exists


class SolveItem(MznItem):
    def __init__(self, name, parent, spec, solver, cache, markers, suite):
        super().__init__(name, parent, solver, markers, suite)
        self.spec = spec
        self.solver = solver
        self.cache = cache
        self.default_options = suite.options
        self.strict = suite.strict

    def runtest(self):
        model, result, required, obtained = self.spec.run(
            str(self.fspath), self.solver, default_options=self.default_options
        )

        # To pass model and result to checker test item
        self.cache.model = model
        self.cache.result = result
        self.cache.obtained = obtained

        passed = self.spec.passed(result)

        if not passed:
            # Test fails if we still haven't passed
            expected = [yaml.dump(exp) for exp in required]
            actual = yaml.dump(obtained)
            self.user_properties.append(("compare", (expected, actual)))
            message = "expected one of\n\n{}\n\nbut got\n\n{}".format(
                "\n---\n".join(expected), actual
            )

            # Doesn't match, so backup by checking against another solver
            if isinstance(result, mzn.Result) and result.status.has_solution():
                checkers = [s for s in self.spec.solvers if s is not self.solver]
                if len(checkers) > 0:
                    checker = checkers[0]
                    non_strict_pass = self.cache.test(checker)
                    status = "but passed" if non_strict_pass else "and failed"
                    message += "\n\n{} check against {}.".format(status, checker)

                    if not self.strict and non_strict_pass:
                        print(message, file=sys.stderr)
                        return

            assert False, message

    def reportinfo(self):
        return self.fspath, 0, "{}::{}".format(str(self.fspath), self.name)


class CheckItem(MznItem):
    def __init__(self, name, parent, cache, solver, checker, markers, suite):
        super().__init__(name, parent, solver, markers, suite)
        if not self.solver_allowed(checker):
            self.add_marker(
                pytest.mark.skip(
                    "skipping checker {} not in {}".format(checker, self.allowed)
                )
            )
        if not self.solver_exists(checker):
            self.add_marker(
                pytest.mark.skip("skipping checker {} not available".format(checker))
            )

        self.cache = cache
        self.solver = solver
        self.checker = checker
        self.user_properties.append(("checker", checker))
        self.add_marker(pytest.mark.check)

    def runtest(self):
        if (
            not isinstance(self.cache.result, mzn.Result)
            or not self.cache.result.status.has_solution()
        ):
            pytest.skip("skipping check for no result/solution")
        else:
            passed = self.cache.test(self.checker)
            assert passed, "failed when checking against {}. Got {}".format(
                self.checker, yaml.dump(self.cache.obtained)
            )

    def reportinfo(self):
        return self.fspath, 0, "{}::{}".format(str(self.fspath), self.name)



from minizinc import default_driver, Driver
import subprocess


def test_fzn_flags_arg_conflict():
    assert isinstance(default_driver, Driver)
    # Should fail with no model file given, and not print help
    p = subprocess.run(
        [default_driver._executable, "--fzn-flags", "--help"],
        stdin=None,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    assert p.returncode == 1


#!/usr/bin/env python3

##  Author: Gleb Belov@monash.edu  2017
##  This program runs MiniZinc solvers over a set of instances,
##  checks solutions and compares to given solution logs.

##  TODO type errors etc. when checking?
##  TODO continuous output dumping as option
##  TODO CPU/user time limit, proper memory limit (setrlimit not working)

import sys, io, re as regex, traceback
import os.path, platform
##import numpy
import math
import json, argparse
import datetime
from collections import OrderedDict

import utils, json_config, json_log, mzn_exec, cmp_result_logs
from json_config import s_CommentKey, s_AddKey

s_ProgramDescr = 'MiniZinc testing automation. (c) 2018 Monash University, gleb.belov@monash.edu'
s_ProgramDescrLong = ( "Allows checking of MiniZinc solutions by configurable checker profiles. The solutions can be input from previous logs or produced by a chosen solver profile, for 1 or several instances, with result comparison and some ranking. New solutions' summary logs, detailed stdout/err outputs, and statistics are saved in subfolder mzn-test/LOGS, /OUTPUTS, and /STATS, resp.")

###########################   GENERAL CONFIG. Could be in the JSON config actually    #########################
sDirResults = "mzn-test"
sFlnSolUnchk = sDirResults + "/sol__unchk.json"      ### Logfile to append immediate results
sFlnSolCheckBase = sDirResults + "/LOGS/sol__{}.json"        ### Logfile to append checked results. {} replaced by datetime
sFlnSolFailBase = sDirResults + "/LOGS/sol__{}__FAIL.json"      ### To save failed solutions
sFlnSolLastDzn = sDirResults + "/sol_last.dzn"     ### File to save the DZN solution for checking
sFlnStdoutBase = sDirResults + "/OUTPUTS/{}.stdout.txt"      ### File name base to dump stdout for any backend call
sFlnStderrBase = sDirResults + "/OUTPUTS/{}.stderr.txt"
sFlnStatLog = sDirResults + "/STATS/stat_log__{}.txt"              ### The instance list name(s) will be inserted, if any

sDZNOutputAgrs = "--output-mode dzn --output-objective"    ## The flattener arguments to produce DZN-compatible output facilitating solution checking
                             ## Remove --output-objective in MZN Release <=2.1.7
sFlatOptChecker = "--allow-multiple-assignments -Dmzn_ignore_symmetry_breaking_constraints=true"

s_UsageExamples = ( 
    "\nUSAGE EXAMPLES:"
    "\n(1)  \"mzn-test.py model.mzn data.dzn [--checkDZN stdout.txt [--checkStderr stderr.txt]] [--chkPrf MINIZINC-CHK --chkPrf FZN-GECODE-CHK] [--tCheck 15] [--addSolverOption \"--fzn-flags '-D fPureCircuit=true'\"]\"                  ::: check the instance's solutions, optionally reading them from a DZN-formatted file (otherwise solving first), optionally overriding default checker list etc."
    "\n(2)  \"mzn-test.py --solver CPLEX -t 300 -l instList1.txt -l instList2.txt --name CPLEXTest_003 --result newLog00.json prevLog1.json prevLog2.json --failed failLog.json\""
    "             ::: solve instances using the specified solver profile and wall time limit 300 seconds. The instances are taken from the list files. The test is aliased CPLEXTest_003. Results are saved to newLog00.json and compared/ranked to those in prevLog's. (Probably) incorrect solutions are saved to failLog.json."
    "\n(3)  \"mzn-test.py [-l instList1.txt] -c prevLog1.json -c prevLog2.json [--runAndCmp]\"                  ::: compare existing logs, optionally limited to the given instances, optionally running new tests. USE SINGLE QUOTES ONLY INSIDE ARGUMENTS PASSED TO THE BACKENDS when running backends through shell."
  )
##############################################################################################
################ Parameters of MZN-Test, including config and command-line
##############################################################################################
class MZT_Param:
    def parseCmdLine(self):
        parser = argparse.ArgumentParser(
          description=s_ProgramDescr + '\n' + s_ProgramDescrLong,
          epilog=s_UsageExamples)
        parser.add_argument('instanceFiles', nargs='*', metavar='<instanceFile>',
          help='model instance files, if no instance lists supplied, otherwise existing solution logs to compare with')
        parser.add_argument('-l', '--instanceList', dest='l_InstLists', action='append', metavar='<instanceListFile>',
            help='file with a list of instance input files, one instance per line,'
              ' instance file types specified in config. Used for running tests or for instance selection in comparison mode')
        parser.add_argument('--name', '--testName', metavar='<string>', help='alias of this test run, defaults to result log file name')
        parser.add_argument('-c', '--compare', action="append", metavar='<logfile>',
                            help='summarize and compare results to existing <logfile>. Only compares the logs and does not run tests, unless --runAndCmp. The flags -c can be omitted if -l is used')
        parser.add_argument('--runAndCmp', '--runAndCompare', '--run', action='store_true',
                            help='even if other logs are provided by -c, do run the tests and compare')
        parser.add_argument('--solver', '--solverPrf', '--solverProfile', metavar='<prf_name>',
                            help='solver profile from those defined in config section \"SOLVER_PROFILES\"')
        parser.add_argument('--call', '--solverCall', metavar='"<exe+flags or shell command(s) if --shellSolve 1>"',
                            help='solver backend call, should be quoted. Insert %%s where instance files need to be. Flatten with \''
                             + sDZNOutputAgrs + '\' to enable solution checking, unless the model has a suitable output definition.'
                             " Pass '--output-time' to the output filter (e.g., solns2out) to enable ranking by time")
        parser.add_argument('-t', '--tSolve',
                            type=float,
                            metavar='<sec>', help='solver backend wall-time limit, default: '+
                              str(self.cfgDefault["BACKEND_DEFS"]["__BE_SOLVER"]["EXE"]["n_TimeoutRealHard"][0]))
        parser.add_argument('--result', default=sFlnSolCheckBase, metavar='<file>',
                            help='save result log to <file>, default: \''+sFlnSolCheckBase+'\'')
        parser.add_argument('--resultUnchk', default=sFlnSolUnchk, metavar='<file>', help='save unchecked result log to <file>')
        parser.add_argument('--chkPrf', '--checkerPrf', '--checkerProfile', metavar='<prf_name>', action='append',
                            help='checker profile from those defined in config section \"CHECKER_PROFILES\", can be a few')
        parser.add_argument('--tCheck',
                            type=float,
                            metavar='<sec>', help='checker backend wall-time limit, default: '+
                              str(self.cfgDefault["BACKEND_DEFS"]["__BE_CHECKER"]["EXE"]["n_TimeoutRealHard"][0]))
        parser.add_argument('--nCheckMax', '--nCheck', '--nCheckedMax', ## default=self.cfgDefault["COMMON_OPTIONS"]["SOLUTION_CHECKING"]["s_FailedSaveFile"][0],
                            type=int,
                            metavar='<N>', help='max number of solutions checked per instance.'
                              ' Negative means checking starts from the last obtained solution. Default: '+
                              str(self.cfgDefault["COMMON_OPTIONS"]["SOLUTION_CHECKING"]["n_CheckedMax"][0]))
        parser.add_argument('--nFailedSaveMax', ## default=self.cfgDefault["COMMON_OPTIONS"]["SOLUTION_CHECKING"]["s_FailedSaveFile"][0],
                            type=int,
                            metavar='<N>', help='max number of failed solution reports saved per instance, default: '+
                              str(self.cfgDefault["COMMON_OPTIONS"]["SOLUTION_CHECKING"]["n_FailedSaveMax"][0]))
        parser.add_argument('--failed', ## default=self.cfgDefault["COMMON_OPTIONS"]["SOLUTION_CHECKING"]["s_FailedSaveFile"][0],
                            metavar='<file>', help='save failed check reports to <file>, default: \''+
                              self.cfgDefault["COMMON_OPTIONS"]["SOLUTION_CHECKING"]["s_FailedSaveFile"][0]+'\'')
        parser.add_argument('--checkDZN', '--checkStdout', metavar='<stdout_file>',
                            help='for a single instance, check DZN-formatted solutions from a solver\'s std output dumped to <stdout_file>. The DZN format is produced, e.g., if the model is flattened with \'' + sDZNOutputAgrs + '\'')
        parser.add_argument('--checkStderr', metavar='<stderr_file>',
                            help='with --checkDZN, read a solver\'s stderr log from <stderr_file> (not essential)')
        parser.add_argument('-v', '--verbose', dest='verbose', action='store_true', help='tee backend\'s stderr to screen, in addition to the instance\'s output dumpfile in mzn-test/OUTPUTS. Only works for --shellSolve 1')
        parser.add_argument('--vc', '--verbose-check', dest='vc', action='store_true', help='same for checking')
     ## parser.add_argument('--no-feature', dest='feature', action='store_false')
     ## parser.set_defaults(feature=True)
        parser.add_argument('--debug', '--printcall', type=int, metavar='<bitfield>', help='bit 1: print full solver call commands, bit 2: same for checker')
        parser.add_argument('--shellSolve', type=int, metavar='0/1', help='backend call through shell when using psutils')
        parser.add_argument('--psutils', type=int, metavar='0/1', help='backend call through psutils (seems buggy in 3.4.2)')
        ## parser.add_argument('--fullPaths', action='store_true',
        ##                    help='use full paths in instance identifiers. By default, it\'s the pure base filenames')
        
        parser.add_argument('--mergeCfg', action="append", metavar='<file>', help='merge config from <file>')
        parser.add_argument('--saveCfg', metavar='<file>', help='save internal config to <file>. Can be useful to modify some parameters and run with --mergeCfg')
        parser.add_argument('--saveSolverCfg', metavar='<file>', help='save the final solver backend config to <file>')
        parser.add_argument('--saveCheckerCfg', metavar='<file>', help='save the final checker backend config to <file>')
        parser.add_argument('--addOption', '--addOptions', action="append", metavar='<text>', type=str, help='add <text> to any solver / checker call')
        parser.add_argument('--addSolverOption', '--addSolverOptions', action="append", metavar='<text>', type=str, help='add <text> to the solver call')
        parser.add_argument('--addCheckerOption', '--addCheckerOptions', action="append", metavar='<text>', type=str, help='add <text> to a checker call')
        parser.add_argument('--useJoinedName', action="append", metavar='<...%s...>', type=str, help='add this to the call, with %%s being replaced by'
                            ' a joined filename from all the input filenames, e.g., "--writeModel MODELS/%%s.mps"')
        self.args = parser.parse_args()
        # print( "ARGS:\n", self.args )
        ## Solver backend and checker backend list
        self.slvBE = None
        self.chkBEs = None

    ## Get parameters from config and command line, merge values
    def obtainParams(self):
        self.parseCmdLine()
        self.mergeValues()
        
    def initCfgDefault(self):
        ddd = {
          s_CommentKey: [
            "The default config structure for mzn-test.py. ", s_ProgramDescr,
            "You can export this by --saveCfg and modify -> --mergeCfg,",
            "even having only partial JSON subtree in a merged file(s).",
            "Structure: COMMON_OPTIONS, SOLVER_/CHECKER_PROFILES, BACKEND_DEFS.",
            "Solver and checkers are selected from pre-defined profiles,",
            "which are in turn built from sequences of 'basic backend definitions'.",
            "Comments are either separate keys or added as list elements ('/// ...')",
            "in pre-selected positions; then, overriding items should keep that order."
          ],
          "COMMON_OPTIONS": {
            s_CommentKey: [ "'Solvers' and 'Checkers' select the profiles to use for solving and checking.",
                            "At the moment only the 1st solver is used from Solvers."
                            "The selected profiles must be known in SOLVER_PROFILES / CHECKER_PROFILES, resp."
                          ],
            "Solvers": [ "MINIZINC",
                        #"MZN-CPLEX",
                        "/// At the moment only the 1st element is used for solving" ],
            "SOLUTION_CHECKING": {
              "Checkers": ["GECODE-CHK", "GUROBI-CHK", "CPLEX-CHK", "ORTOOLS-CHK" ],
              "n_CheckedMax": [ -10, "/// Negative value means it's that many last solutions" ],
              "n_FailedSaveMax": [ 3, "/// After that many failed solutions, stop checking the instance" ],
              "s_FailedSaveFile": [ sFlnSolFailBase, "/// Filename to save failed solutions" ],
            },
            "Instance_List": {
              s_CommentKey: [ "Params for instance lists.",
                              "Instance list is a file containing an instance's model files,",
                              "at most 1 instance per line.",
                              "InstanceFileExt: only files with this extensions from a list file",
                              "will be taken on each line" ],
              "InstanceFileExt": [".mzn", ".dzn"]     ##   Add json?   TODO
            },
            "runCommand": {
              "windows": {
                "runSilent": "echo \"WARNING. No timeout on Windows.\" & {2} 1>{3} 2>{4}",
                "runVerbose": "echo \"WARNING. No timeout on Windows.\" & {2} 3>&1 1>{3} 2>&3 | tee {4} & echo >>{4}"
              }
              ,"non-windows": {
                "runSilent": "ulimit -v {0}; timeout -k 1 {1} bash -c \"{2}\" 1>{3} 2>{4}",
                "runVerbose": "ulimit -v {0}; timeout -k 1 {1} bash -c \"{2}\" 3>&1 1>{3} 2>&3 | tee {4}; echo >>{4}"
              }
            }
          },
          "SOLVER_PROFILES": {
            s_CommentKey: [ "Similar to CHECKER_PROFILES." ],
            "MINIZINC": [ "__BE_COMMON", "__BE_SOLVER", "BE_MINIZINC" ],
            "FZN-GUROBI": [ "__BE_COMMON", "__BE_SOLVER", "BE_FZN-GUROBI" ],
            "FZN-CPLEX": [ "__BE_COMMON", "__BE_SOLVER", "BE_FZN-CPLEX" ],
            "FZN-CBC": [ "__BE_COMMON", "__BE_SOLVER", "BE_FZN-CBC" ],
            "GUROBI": [ "__BE_COMMON", "__BE_SOLVER", "BE_GUROBI" ],
            "CPLEX": [ "__BE_COMMON", "__BE_SOLVER", "BE_CPLEX" ],
            "XPRESS": [ "__BE_COMMON", "__BE_SOLVER", "BE_XPRESS" ],
            "SCIP": [ "__BE_COMMON", "__BE_SOLVER", "BE_SCIP" ],
            "CBC": [ "__BE_COMMON", "__BE_SOLVER", "BE_CBC" ],
            "GECODE": [ "__BE_COMMON", "__BE_SOLVER", "BE_GECODE" ],
            "CHUFFED": [ "__BE_COMMON", "__BE_SOLVER", "BE_CHUFFED" ],
            "ORTOOLS": [ "__BE_COMMON", "__BE_SOLVER", "BE_ORTOOLS" ] #,
            # "MZN-GUROBI": [ "__BE_COMMON", "__BE_SOLVER", "BE_MZN-GUROBI" ],
            # "MZN-CPLEX": [ "__BE_COMMON", "__BE_SOLVER", "BE_MZN-CPLEX" ],
            # "MZN-CBC": [ "__BE_COMMON", "__BE_SOLVER", "BE_MZN-CBC" ],
            # "MZN-GECODE": [ "__BE_COMMON", "__BE_SOLVER", "BE_MZN-GECODE" ],
            # "FZN-CHUFFED": [ "__BE_COMMON", "__BE_SOLVER", "BE_FZN-CHUFFED" ]
          },
          "CHECKER_PROFILES": {
            s_CommentKey: [ "Each profile gives a list of backend defs to use.",
              "Later backends in the list can override/add options, ",
              "for example for values to be read from the outputs.",
              "Adding is only possible if the key is prefixed by '"+s_AddKey+"'"
            ],
            "MINIZINC-CHK": [ "__BE_COMMON", "__BE_CHECKER_OLDMINIZINC", "BE_MINIZINC" ],
            "GECODE-CHK": [ "__BE_COMMON", "__BE_CHECKER", "BE_GECODE" ],
            "ORTOOLS-CHK": [ "__BE_COMMON", "__BE_CHECKER", "BE_ORTOOLS" ],
            "GUROBI-CHK": [ "__BE_COMMON", "__BE_CHECKER", "BE_GUROBI" ],
            "CPLEX-CHK": [ "__BE_COMMON", "__BE_CHECKER", "BE_CPLEX" ],
            "CHUFFED-CHK": [ "__BE_COMMON", "__BE_CHECKER", "BE_CHUFFED" ],
            "MZN-GECODE-CHK": [ "__BE_COMMON", "__BE_CHECKER", "BE_MZN-GECODE" ],
            "MZN-GUROBI-CHK": [ "__BE_COMMON", "__BE_CHECKER", "BE_MZN-GUROBI" ],
            "MZN-CPLEX-CHK": [ "__BE_COMMON", "__BE_CHECKER", "BE_MZN-CPLEX" ],
            "FZN-GECODE-CHK": [ "__BE_COMMON", "__BE_CHECKER_OLDMINIZINC", "BE_FZN-GECODE" ],
            "FZN-GECODE-SHELL-CHK": [ "__BE_COMMON", "__BE_CHECKER_OLDMINIZINC", "BE_FZN-GECODE_SHELL" ],
            "FZN-CHUFFED-CHK": [ "__BE_COMMON", "__BE_CHECKER", "BE_FZN-CHUFFED" ]
          },
          "BACKEND_DEFS": {
            s_CommentKey: [ "__BE_COMMON initializes a basic backend structure.",
              "Each further backend in a profile list overrides or adds options."
            ],
            "__BE_COMMON": {
              s_CommentKey: [ "THE INITIALIZING BACKEND." ],
              "EXE": {
                s_CommentKey: [ "Solver call parameters" ],
                "s_SolverCall" : ["minizinc -v -s -i " + sDZNOutputAgrs + " %s",
                  "/// The 1st element defines the call line. %s is replaced by the instance filename(s)."],
                "s_ExtraCmdline" : ["", "/// Only for __BE_SOLVER/__BE_CHECKER... subprofiles."
                                        " The 1st element gives extra cmdline arguments to the call"],
                "b_ThruShell"  : [True, "/// Set True to call solver thru shell."
                  " Then you can do shell tricks but Ctrl+C may not kill all subprocesses etc."],
                "n_TimeoutRealHard": [150, "/// Real-time timeout per instance, seconds,"
                  " for all solution steps together. Use mzn/backend options for CPU time limit."],
                "n_VMEMLIMIT_SoftHard": [8000000, 8000000, "/// 2 limits, soft/hard, in KB. Platform-dependent in Python 3.6. Default 8 GB"],
              },
              "Stderr_Keylines": {
                s_CommentKey: [ "A complete line in stderr will be interpreted accordingly.",
                  " Format: <outvar> : { <line>: <value>, ... }"
                  " You can add own things here (use '"+s_AddKey+"' before new var name)",
                  " which will be transferred into results" ]
              },
              "Stderr_Keyvalues": {
                s_CommentKey: [ "Numerical values to be extracted from a line in stderr.",
                  " { <outvar>: [ <regex search pattern>, <regex to replace by spaces>, <value's pos in the line>] }."
                ], 
                ### The %%%mzn-stat values appear in stdout (as of May 2019) but leave them here just in case
                "Time_Flt": [ "%%%mzn-stat: flatTime", "[:=]", 3, "/// E.g., 'Flattening done, 3s' produces 3."
                                " !!! This is interpreted as successful flattening by the checker" ],
                "ObjVal_Solver":   [ "%%%mzn-stat: objective=", "[,:/=]", 3,        ## Need = to avoid mixup with the bound
                                        "/// The objval as reported by solver."],
                "DualBnd_Solver":   [ "%%%mzn-stat: objectiveBound", "[,:/=]", 3 ],
                "CPUTime_Solver":   [ "%%%mzn-stat: solveTime", "[,:/=]", 3 ],
                "NNodes_Solver":   [ "%%%mzn-stat: nodes", "[,:/=]", 3 ],
              },
              "Stdout_Keylines": {
                s_CommentKey: [ "Similar to Stderr_Keylines"],
                "Sol_Status": {
                  "----------": 1,
                  "==========": 2,
                  "=====UNSATISFIABLE=====": -1,
                  "=====UNBOUNDED=====": -2,
                  "=====UNKNOWN=====": 0,
                  "=====UNSATorUNBOUNDED=====": -3,
                  "=====ERROR=====": -4
                },
                "Problem_Sense": {
                  "%%%mzn-stat: method=\"maximize\"": 1,
                  "%%%mzn-stat: method=\"minimize\"": -1,
                  "%%%mzn-stat: method=\"satisfy\"": 0,
                }
              },
              "Stdout_Keyvalues": {
                s_CommentKey: ["Similar to Stderr_Keyvalues." ],
                "Time_Flt": [ "%%%mzn-stat: flatTime", "[:=]", 3, "/// E.g., 'Flattening done, 3s' produces 3."
                                " !!! This is interpreted as successful flattening by the checker" ],
                "ObjVal_Solver":   [ "%%%mzn-stat: objective=", "[,:/=]", 3,        ## Need = to avoid mixup with the bound
                                        "/// The objval as reported by solver."],
                "DualBnd_Solver":   [ "%%%mzn-stat: objectiveBound", "[,:/=]", 3 ],
                "CPUTime_Solver":   [ "%%%mzn-stat: solveTime", "[,:/=]", 3 ],
                "NNodes_Solver":   [ "%%%mzn-stat: nodes", "[,:/=]", 3 ],
                "ObjVal_MZN":   [ "_objective", "[():=;%]", 2,
                                    "/// The objective value as evaluated by MZN." ],
                "RealTime_Solns2Out": [ "% time elapsed:", " ", 4 ],
              }
            },
            "__BE_SOLVER": {
              s_CommentKey: ["Specializations for a general solver" ],
              "EXE": {
                "s_ExtraCmdline" : ["-i"],
                "b_ThruShell"  : [True],
                "n_TimeoutRealHard": [150],
                #  "n_VMEMLIMIT_SoftHard": [8000100000, 8100000000]
              }
            },
            "__BE_CHECKER": {
              s_CommentKey: ["Specializations for a general checker" ],
              "EXE": {
                "s_ExtraCmdline" : [sFlatOptChecker],
                "b_ThruShell"  : [True],
                "n_TimeoutRealHard": [15],
                #  "n_VMEMLIMIT_SoftHard": [8100000000, 8100000000]
              }
            },
            "__BE_CHECKER_OLDMINIZINC": {
              s_CommentKey: ["Specializations for a general checker using the 1.6 MiniZinc driver" ],
              "EXE": {
                "s_ExtraCmdline" : ["--mzn2fzn-cmd 'mzn2fzn -v -s --output-mode dzn " + sFlatOptChecker + "'"],
                "b_ThruShell"  : [True],
                "n_TimeoutRealHard": [15],
                #  "n_VMEMLIMIT_SoftHard": [8100000000, 8100000000]
              }
            },
            "BE_MINIZINC": {
              s_CommentKey: [ "------------------- Specializations for pure minizinc driver" ],
              "EXE":{
                "s_SolverCall": [ "minizinc --mzn2fzn-cmd 'mzn2fzn -v -s " + sDZNOutputAgrs + "' -s %s"], # _objective fails for checking
                "b_ThruShell"  : [True],
              },
            },
            "BE_FZN-GUROBI": {
              s_CommentKey: [ "------------------- Specializations for Gurobi solver instance" ],
              "EXE":{
                "s_SolverCall" : ["mzn2fzn -v -s -G linear " + sDZNOutputAgrs + " %s --fzn tmp.fzn --ozn tmp.ozn && mzn-gurobi -v -s tmp.fzn"],  ## works without solns2out for now. Need thus when using shell call with system()  TODO?
                #"opt_writeModel": ["--writeModel"]
              },
              "Stderr_Keyvalues": {
                s_AddKey+"Preslv_Rows": [ "Presolved:", " ", 2 ],
                s_AddKey+"Preslv_Cols": [ "Presolved:", " ", 4 ],
                s_AddKey+"Preslv_Non0": [ "Presolved:", " ", 6 ]
              },
            },
            "BE_GUROBI": {
              s_CommentKey: [ "------------------- Specializations for Gurobi solver instance" ],
              "EXE":{
                "s_SolverCall" : ["minizinc -v -s --solver gurobi --output-time " + sDZNOutputAgrs + " %s"], # _objective fails for checking TODO
                #"opt_writeModel": ["--writeModel"]
              },
              "Stderr_Keyvalues": {
                s_AddKey+"Preslv_Rows": [ "Presolved:", " ", 2 ],
                s_AddKey+"Preslv_Cols": [ "Presolved:", " ", 4 ],
                s_AddKey+"Preslv_Non0": [ "Presolved:", " ", 6 ]
              },
            },
            "BE_MZN-GUROBI": {
              s_CommentKey: [ "------------------- Specializations for Gurobi solver instance" ],
              "EXE":{
                "s_SolverCall" : ["mzn-gurobi -v -s -G linear --output-time " + sDZNOutputAgrs + " %s"], # _objective fails for checking TODO
                #"opt_writeModel": ["--writeModel"]
              },
              "Stderr_Keyvalues": {
                s_AddKey+"Preslv_Rows": [ "Presolved:", " ", 2 ],
                s_AddKey+"Preslv_Cols": [ "Presolved:", " ", 4 ],
                s_AddKey+"Preslv_Non0": [ "Presolved:", " ", 6 ]
              },
            },
            "BE_FZN-CPLEX": {
              s_CommentKey: [ "------------------- Specializations for IBM ILOG CPLEX solver instance" ],
              "EXE": {
                "s_SolverCall" : ["mzn2fzn -v -s -G linear " + sDZNOutputAgrs + " %s --fzn tmp.fzn --ozn tmp.ozn && mzn-cplex -v -s tmp.fzn"]
                #"s_SolverCall" : ["./run-mzn-cplex.sh %s"],
                #"b_ThruShell"  : [True],
                #"opt_writeModel": ["--writeModel"]
              },
              "Stderr_Keyvalues": {
                s_AddKey+"Preslv_Rows": [ "Reduced MIP has [0-9]+ rows,", " ", 4 ],
                s_AddKey+"Preslv_Cols": [ "Reduced MIP has [0-9]+ rows,", " ", 6 ],
                s_AddKey+"Preslv_Non0": [ "Reduced MIP has [0-9]+ rows,", " ", 9 ]
              },
            },
            "BE_CPLEX": {
              s_CommentKey: [ "------------------- Specializations for IBM ILOG CPLEX solver instance" ],
              "EXE": {
                "s_SolverCall" : ["minizinc -v -s --solver cplex --output-time " + sDZNOutputAgrs + " %s"], # _objective fails for checking
                #"s_SolverCall" : ["./run-mzn-cplex.sh %s"],
                #"b_ThruShell"  : [True],
                #"opt_writeModel": ["--writeModel"]
              },
              "Stderr_Keyvalues": {
                s_AddKey+"Preslv_Rows": [ "Reduced MIP has [0-9]+ rows,", " ", 4 ],
                s_AddKey+"Preslv_Cols": [ "Reduced MIP has [0-9]+ rows,", " ", 6 ],
                s_AddKey+"Preslv_Non0": [ "Reduced MIP has [0-9]+ rows,", " ", 9 ]
              },
            },
            "BE_XPRESS": {
              s_CommentKey: [ "------------------- Specializations for FICO XPRESS solver instance" ],
              "EXE": {
                "s_SolverCall" : ["minizinc -v -s --solver xpress --output-time " + sDZNOutputAgrs + " %s"], # _objective fails for checking
                #"s_SolverCall" : ["./run-mzn-cplex.sh %s"],
                #"b_ThruShell"  : [True],
                #"opt_writeModel": ["--writeModel"]
              },
              "Stderr_Keyvalues": {   ## Is different for XPRESS and -v fails at the moment anyway
                s_AddKey+"Preslv_Rows": [ "Reduced MIP has [0-9]+ rows,", " ", 4 ],
                s_AddKey+"Preslv_Cols": [ "Reduced MIP has [0-9]+ rows,", " ", 6 ],
                s_AddKey+"Preslv_Non0": [ "Reduced MIP has [0-9]+ rows,", " ", 9 ]
              },
            },
            "BE_MZN-CPLEX": {
              s_CommentKey: [ "------------------- Specializations for IBM ILOG CPLEX solver instance" ],
              "EXE": {
                "s_SolverCall" : ["mzn-cplex -v -s -G linear --output-time " + sDZNOutputAgrs + " %s"], # _objective fails for checking
                #"s_SolverCall" : ["./run-mzn-cplex.sh %s"],
                #"b_ThruShell"  : [True],
                #"opt_writeModel": ["--writeModel"]
              },
              "Stderr_Keyvalues": {
                s_AddKey+"Preslv_Rows": [ "Reduced MIP has [0-9]+ rows,", " ", 4 ],
                s_AddKey+"Preslv_Cols": [ "Reduced MIP has [0-9]+ rows,", " ", 6 ],
                s_AddKey+"Preslv_Non0": [ "Reduced MIP has [0-9]+ rows,", " ", 9 ]
              },
            },
            "BE_FZN-CBC": {
              s_CommentKey: [ "------------------- Specializations for COIN-OR Branch&Cut solver instance" ],
              "EXE": {
                "s_SolverCall" : ["mzn-cbc -v -s -G linear --output-time " + sDZNOutputAgrs + " %s --fzn tmp.fzn --ozn tmp.ozn && mzn-cplex -v -s tmp.fzn"], 
                #"s_SolverCall" : ["./run-mzn-cplex.sh %s"],
                #"b_ThruShell"  : [True],
              },
            },
            "BE_SCIP": {
              s_CommentKey: [ "------------------- Specializations for SCIP solver instance" ],
              "EXE": {
                "s_SolverCall" : ["minizinc -v -s --solver scip --output-time " + sDZNOutputAgrs + " %s"], # _objective fails for checking
                #"s_SolverCall" : ["./run-mzn-cplex.sh %s"],
                #"b_ThruShell"  : [True],
              },
            },
            "BE_CBC": {
              s_CommentKey: [ "------------------- Specializations for COIN-OR Branch&Cut solver instance" ],
              "EXE": {
                "s_SolverCall" : ["minizinc -v -s --solver osicbc --output-time " + sDZNOutputAgrs + " %s"], # _objective fails for checking
                #"s_SolverCall" : ["./run-mzn-cplex.sh %s"],
                #"b_ThruShell"  : [True],
              },
            },
            "BE_MZN-CBC": {
              s_CommentKey: [ "------------------- Specializations for COIN-OR Branch&Cut solver instance" ],
              "EXE": {
                "s_SolverCall" : ["mzn-cbc -v -s -G linear --output-time " + sDZNOutputAgrs + " %s"], # _objective fails for checking
                #"s_SolverCall" : ["./run-mzn-cplex.sh %s"],
                #"b_ThruShell"  : [True],
              },
            },
            "BE_GECODE": {
              s_CommentKey: [ "------------------- Specializations for Gecode FlatZinc interpreter" ],
              "EXE": {
#                "s_SolverCall" : ["minizinc -s --solver gecode " + sDZNOutputAgrs + " %s"], # _objective fails for checking TODO
                "s_SolverCall" : ["minizinc -v -s --solver gecode " + sDZNOutputAgrs
                     + " %s"],    #  --time 300000
                "b_ThruShell"  : [True],
              }
            },
            "BE_MZN-GECODE": {
              s_CommentKey: [ "------------------- Specializations for Gecode FlatZinc interpreter" ],
              "EXE": {
#                "s_SolverCall" : ["mzn-fzn -s -G gecode --solver fzn-gecode " + sDZNOutputAgrs + " %s"], # _objective fails for checking TODO
                "s_SolverCall" : ["mzn-gecode -v -s -G gecode " + sDZNOutputAgrs
                     + " %s"],    #  --time 300000
                "b_ThruShell"  : [True],
              }
            },
            "BE_FZN-GECODE": {
              s_CommentKey: [ "------------------- Specializations for Gecode FlatZinc interpreter" ],
              "EXE": {
#                "s_SolverCall" : ["mzn-fzn -s -G gecode --solver fzn-gecode " + sDZNOutputAgrs + " %s"], # _objective fails for checking TODO
                "s_SolverCall" : ["minizinc -s -G gecode -f fzn-gecode --mzn2fzn-cmd 'mzn2fzn -v -s " + sDZNOutputAgrs + "' %s"],
                "b_ThruShell"  : [True],
              }
            },
            "BE_FZN-GECODE_SHELL": {
              s_CommentKey: [ "------------------- Specializations for Gecode FlatZinc interpreter" ],
              "EXE": {
#                "s_SolverCall" : ["mzn-fzn -s -G gecode --solver fzn-gecode " + sDZNOutputAgrs + " %s"], # _objective fails for checking TODO
                "s_SolverCall" : ["mzn2fzn -v -s -G gecode " + sDZNOutputAgrs
                     + sFlatOptChecker + " %s --fzn tmp.fzn --ozn tmp.ozn && fzn-gecode tmp.fzn | solns2out tmp.ozn"],
                "b_ThruShell"  : [True],
                "s_ExtraCmdline" : [""],
              }
            },
            "BE_CHUFFED": {
              s_CommentKey: [ "------------------- Specializations for Chuffed FlatZinc interpreter" ],
              "EXE": {
                "s_SolverCall" : ["minizinc -v -s --solver chuffed -f --output-time "
                                    + sDZNOutputAgrs + " %s"], # _objective fails for checking
              }  ##  --fzn-flags --time-out --fzn-flags 300
            }
            , "BE_ORTOOLS": {
              s_CommentKey: [ "------------------- Specializations for OR-Tools FlatZinc interpreter" ],
              "EXE": {
                "s_SolverCall" : ["minizinc -v -s --solver ortools -f --output-time "
                                    + sDZNOutputAgrs + " %s"], # _objective fails for checking
              }  ##  --fzn-flags --time-out --fzn-flags 300
            }
            , "BE_FZN-CHUFFED": {
              s_CommentKey: [ "------------------- Specializations for Chuffed FlatZinc interpreter" ],
              "EXE": {
                "s_SolverCall" : ["mzn-fzn -v -s -G chuffed --solver fzn-chuffed --fzn-flags -f --output-time "
                                    + sDZNOutputAgrs + " %s"], # _objective fails for checking
              }  ##  --fzn-flags --time-out --fzn-flags 300
            }
          }
        }
        return ddd
    ##        self.nNoOptAndAtLeast2Feas = 0
    
    ## Read a cfg file, instead of/in addition to the default cfg
    def mergeCfg(self, fln):
        ddd1 = None
        with open( fln, 'r' ) as rf:
            ddd1 = json.load( rf )
        self.cfg = mergeJSON( self.cfg, ddd1 )
      
    ## Merge cmdline values with cfg, performing some immediate actions
    ## And compile the backends constituting solver and checker(s)
    def mergeValues(self):
        if None!=self.args.mergeCfg:               ### MERGE CFG FROM EXTRA FILES
            for eCfg in self.args.mergeCfg:
                self.mergeCfg( eCfg )
        ################ Update some explicit cmdline params -- AFTER MERGING CFG FILES.
        if None!=self.args.failed:
            self.cfg["COMMON_OPTIONS"]["SOLUTION_CHECKING"]["s_FailedSaveFile"][0] = self.args.failed
        if None!=self.args.nCheckMax:
            self.cfg["COMMON_OPTIONS"]["SOLUTION_CHECKING"]["n_CheckedMax"][0] = self.args.nCheckMax
        if None!=self.args.nFailedSaveMax:
            self.cfg["COMMON_OPTIONS"]["SOLUTION_CHECKING"]["n_FailedSaveMax"][0] = self.args.nFailedSaveMax
        ################ SAVE FINAL CFG
        if None!=self.args.saveCfg:
            with utils.openFile_autoDir( self.args.saveCfg, 'w' ) as wf:
                print( "Saving final config to", self.args.saveCfg )
                json.dump( self.cfg, wf, sort_keys=True, indent=json_config.n_JSON_Indent )
        ### COMPILE THE SOLVER BACKEND
        if None!=self.args.solver:
            self.cfg["COMMON_OPTIONS"]["Solvers"][0] = self.args.solver
        slvPrfName = self.cfg["COMMON_OPTIONS"]["Solvers"][0]
        slvPrf = self.cfg["SOLVER_PROFILES"][slvPrfName]
        assert len(slvPrf)>0, "Solver profile '%s' should use at least a basic backend" % slvPrfName
        self.slvBE = self.cfg["BACKEND_DEFS"][slvPrf[0]]
        for i in range( 1, len( slvPrf ) ):
            self.slvBE = json_config.mergeJSON( self.slvBE, self.cfg["BACKEND_DEFS"][slvPrf[i]] )
        if None!=self.args.tSolve:
            self.slvBE["EXE"]["n_TimeoutRealHard"][0] = self.args.tSolve
        assert None==self.args.solver or None==self.args.call, "ERROR: both solver call and a solver profile specified."
        if None!=self.args.call:    ## After the compilation
            self.slvBE["EXE"]["s_SolverCall"][0] = self.args.call
        if None!=self.args.shellSolve:
            self.slvBE["EXE"]["b_ThruShell"][0] = self.args.shellSolve!=0
        print ( "\nSolver/checker configurations:\n     SLV_CFG: ", json.dumps( self.slvBE["EXE"] ) )
        ### COMPILE THE CHECKER BACKENDS
        if None!=self.args.chkPrf and 0<len( self.args.chkPrf ):
            self.cfg["COMMON_OPTIONS"]["SOLUTION_CHECKING"]["Checkers"] = self.args.chkPrf
        chkPrfList = self.cfg["COMMON_OPTIONS"]["SOLUTION_CHECKING"]["Checkers"]
        self.chkBEs = []
        for chkPrfName in chkPrfList:
            chkPrf = self.cfg["CHECKER_PROFILES"][chkPrfName]
            assert len(chkPrf)>0, "Checker profile '%s' should use at least a basic backend" % chkPrfName
            self.chkBEs.append( self.cfg["BACKEND_DEFS"][chkPrf[0]] )
            for i in range( 1, len( chkPrf ) ):
                self.chkBEs[-1] = json_config.mergeJSON( self.chkBEs[-1], self.cfg["BACKEND_DEFS"][chkPrf[i]] )
            if None!=self.args.tCheck:
                self.chkBEs[-1]["EXE"]["n_TimeoutRealHard"][0] = self.args.tCheck
            print ( "     CHK_CFG: ", json.dumps( self.chkBEs[-1]["EXE"] ) )
        print( "" )
        ### SAVE THE SOLVER BACKEND
        if None!=self.args.saveSolverCfg:
            with utils.openFile_autoDir( self.args.saveSolverCfg, 'w' ) as wf:
                print( "Saving solver config to", self.args.saveSolverCfg )
                json.dump( self.slvBE, wf, sort_keys=True, indent=json_config.n_JSON_Indent )
        ### SAVE THE CHECKER BACKENDS
        if None!=self.args.saveCheckerCfg:
            with utils.openFile_autoDir( self.args.saveCheckerCfg, 'w' ) as wf:
                print( "Saving checker config to", self.args.saveCheckerCfg )
                json.dump( self.chkBE, wf, sort_keys=True, indent=json_config.n_JSON_Indent )
        self.sThisName = self.args.result
        if None!=self.args.name:
            self.sThisName = self.args.name
    
    def __init__(self):
        self.cfgDefault = self.initCfgDefault()
        self.cfg = self.cfgDefault
        ### Further parameters
        self.args = {}
        
    def __str__(self):
        s_Out = json.dumps( self.cfg, sort_keys=True, indent=json_config.n_JSON_Indent ) + '\n'
        s_Out += str(self.args) + '\n'
        s_Out += "\n >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> SOLVER BACKEND:\n"
        s_Out += json.dumps( self.slvBE, sort_keys=True, indent=json_config.n_JSON_Indent ) + '\n'
        s_Out += "\n >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> CHECKER BACKENDS:\n"
        for chkBE in self.chkBEs:
            s_Out += json.dumps( chkBE, sort_keys=True, indent=json_config.n_JSON_Indent ) + '\n'
        return s_Out
                

##############################################################################################
################### The MZNTest class
##############################################################################################
class MznTest:

    ## Produce a pair: first, a string identifying the instance which is given as a string of the instance files
    ## Second, same without paths and file extensions for short printing
    ## the elements are sorted according to the extensions list, then alphabetically
    def getIName( self, s_Inst ):
        lId = s_Inst.split()     ## list of instance files
        lExt = self.params.cfg["COMMON_OPTIONS"]["Instance_List"]["InstanceFileExt"]
        lId = sorted( lId, key = lambda nm:
                ( lExt.index( os.path.splitext(nm)[1] ) \
                        if os.path.splitext(nm)[1] in lExt else len(lExt), os.path.basename( nm ) ) )
        lId2 = [ os.path.splitext( os.path.basename( fln ) )[0] for fln in lId ];
        return ' '.join(lId), ' '.join(lId2);

    def obtainParams( self ):
        self.params.obtainParams()

    ## If an instance was specified in cmdline, or model list(s) supplied
    def compileExplicitModelLists(self):
        ## Get cmdline filenames or from the --instList arguments
        ## Can compile the list from log files, see below
        self.params.instList = []
        ## Only if -l not used, take the pos args
        if (self.params.args.l_InstLists is None or 0==len( self.params.args.l_InstLists )) \
                and self.params.args.instanceFiles is not None and 0<len( self.params.args.instanceFiles ):
            self.params.instList.append( " ".join( self.params.args.instanceFiles ) )
        ## Mode "compare only" if (comparison lists and not run option) or no instances
        self.bCmpOnly = True if (not self.params.args.runAndCmp and (( \
                self.params.args.compare is not None and 0<len(self.params.args.compare)) or \
              ( self.params.args.l_InstLists is not None and 0<len( self.params.args.l_InstLists ) \
              and self.params.args.instanceFiles is not None and 0<len( self.params.args.instanceFiles ) ))) \
          else False
        ## If -l used, compile the inst list files
        if None!=self.params.args.l_InstLists and 0<len( self.params.args.l_InstLists ):
            for sFln in self.params.args.l_InstLists:
                with open( sFln, 'r' ) as rf:
                    for line in rf:
                        s_ModelFiles = ""
                        for wrd in line.split():
                            fIsMF = False
                            for ext in self.params.cfg["COMMON_OPTIONS"]["Instance_List"]["InstanceFileExt"]:
                                if wrd.endswith( ext ):
                                    fIsMF = True
                                    break
                            if fIsMF:
                                s_ModelFiles += ' ' + wrd
                        if 0<len( s_ModelFiles ):
                          self.params.instList.append( s_ModelFiles )
        
    ## Result logs can be used to extract the instance list, if wished
    def compileResultLogs(self):
        self.cmpRes = cmp_result_logs.CompareLogs()
        cmpFileList = []
        if None!=self.params.args.compare:     ### If -c used
            cmpFileList += self.params.args.compare
        ### If -l used, interpret pos arguments as comparison logs
        if None!=self.params.args.l_InstLists and 0<len( self.params.args.l_InstLists ):
            cmpFileList += self.params.args.instanceFiles
        for sFlnLog in cmpFileList:
            nEntries = 0
            logCurrent, lLogNames = self.cmpRes.addLog( sFlnLog )
            print( "Reading result log '", sFlnLog, "'...  ", sep='', end='', flush=True )
            with open( sFlnLog, 'r' ) as rf:
                while True:
                    chJ = json_log.readLogJSONChunk( rf )
                    if None==chJ:
                        break
                    try:
                        logCurrent[ self.getIName( chJ["Inst_Files"] )[0] ] = chJ
                        if "TEST_NAME" in chJ:
                            lLogNames[1] = chJ[ "TEST_NAME" ]
                            print( "  TEST NAME: '", chJ[ "TEST_NAME" ],
                                  "'... ", sep='', end='', flush=True )
                    except:
                        print( "\n   WARNING: unrecognized result chunk in file '", sFlnLog,
                              "' before position", rf.tell(), ", doesn't contain all keys")
                    else:
                        nEntries += 1
            print( len(logCurrent), "different instances among", nEntries, "recognized entries."  )
        
    def runTheInstances(self):
        logCurrent, lLogNames = self.cmpRes.addLog( self.params.args.result )
        if self.params.sThisName!=lLogNames[0]:
            lLogNames[1] = self.params.sThisName
        ## TRUNCATING files first, then "a" - better on Win??
        sNow = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        self.sStartTime = sNow
        self.fileSol00 = utils.openFile_autoDir( self.params.args.resultUnchk, "w" )
        self.fileSol = utils.openFile_autoDir(self.params.args.result.format(sNow + 
          ("" if self.params.args.name is None else "__"+utils.flnfy(self.params.args.name))), "w" )
        self.fileFailName = self.params.cfg["COMMON_OPTIONS"]["SOLUTION_CHECKING"]["s_FailedSaveFile"][0].format(sNow)
        self.fileFail = None      ## Not opening yet
        self.nCheckedInstances = 0
        self.nChecksFailed = 0
        self.cmpRes.initListComparison()
        for i_Inst in range( len(self.params.instList) ):
            s_Inst = self.params.instList[ i_Inst ]
            self.initInstance( i_Inst, s_Inst )
            self.solveOriginal( s_Inst )
            try:
                if self.ifShouldCheck():
                    self.checkOriginal( s_Inst )
                else:
                    print( "NO CHECK,  total check-failed instances:", self.nChecksFailed,
                           "from", self.nCheckedInstances )
            except:
                print( "  WARNING: failed to check instance solution. ", sys.exc_info() )
            ## Saving to the main log:
            self.saveSolution( self.fileSol )
            ## Ranking:
            sSet_Inst = self.getIName( self.result["Inst_Files"] )
            logCurrent[ sSet_Inst[0] ] = self.result
            try:
                self.cmpRes.compareInstance( sSet_Inst )
                if i_Inst < len(self.params.instList)-1:
                    print( "STATS:  ", end='' )
                    self.cmpRes.summarizeCurrent( lLogNames )
                    print( "" )
            except:
                print( "  WARNING: failed to compare/rank instance. ", sys.exc_info() )

    def compareLogs(self):
        self.cmpRes.initListComparison()
        theList = [ lfn for lfn in self.params.instList ]
        if 0==len( theList ):
            theList = self.cmpRes.getInstanceUnion()
        for sInst in theList:
            try:
                self.cmpRes.compareInstance( self.getIName( sInst ) )
            except:
                print( "  ------  WARNING: failed to compare/rank instance. ",  )
                traceback.print_exc()
        self.cmpRes.summarizeCmp()
        print( self.cmpRes.summarizeFinalHdr(), end='' )

    def summarize(self):
        try:
            ### Printing summary
            print ('')
            print ('='*50)
            print('            SUMMARY')
            print ('='*50)
            stats = self.cmpRes.summarize()
            print (stats)
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            print(exc_type, exc_obj, fname, exc_tb.tb_lineno)
        if not self.bCmpOnly:
            print( "Printing stats log to: ", end="" )
            sFlnSL = sFlnStatLog.format( utils.flnfy(
                " ".join(self.params.args.l_InstLists if self.params.args.l_InstLists is not None else []) ) )
            print( sFlnSL )
            with utils.openFile_autoDir( sFlnSL, "a" ) as wf:
                wf.write( "\nRUN  " + self.sStartTime + "--" + datetime.datetime.now().__str__() + ":  " )
                wf.write( sys.argv.__str__() )
                wf.write( "\n" )
                wf.write( stats )
            print( "\nResult logs saved to '",  self.params.args.result,
               "', with the unchecked log in '", self.params.args.resultUnchk, "'; failed solutions saved to '",
               self.params.cfg["COMMON_OPTIONS"]["SOLUTION_CHECKING"]["s_FailedSaveFile"][0],
               "' in an \"appendable JSON\" format, cf. https://github.com/pvorb/jsml."
               ## no more for shell: "\nSolver/checker stdout(err) outputs saved to '" + sDirResults + "/last_stdout(err)_...txt'."
               , sep='')

##############################################################################################
######################### MZNTest average-level #############################

    ## init
    def initInstance( self, i_Inst, s_Inst ):
        print( "\n--------------------- INSTANCE %d of %d:  '" % (i_Inst+1, len( self.params.instList )),
              s_Inst, "' ----------------------", sep='' )
        self.result = OrderedDict()
        if 0==i_Inst and None!=self.params.args.name:
            self.result["TEST_NAME"] = self.params.args.name
        self.result[ "Inst_Index_Of" ] = [ (i_Inst+1), len( self.params.instList ) ]
        self.result["Inst_Files"] = s_Inst
        self.solList = []                   ## The solution list
        ## TODO: instance name source?

    ## Solve the original instance
    def solveOriginal( self, s_Inst ):
        self.result["__SOLVE__"] = self.solveInstance( s_Inst, self.params.slvBE, '_SOLVING', self.solList )
        # print( "   RESULT:\n", self.result )
        self.saveSolution( self.fileSol00 )
        
    ## Solve a given MZN instance and return the result map
    ## Arguments: instance files in a string, backend parameter dictionary
    ## slvName is used for screen output and last_std(out/err)_... naming, so better no spaces
    ## solList provided <=> this is solving (not checking) and will use --checkDZN if opted

    ##  TODO
    ## Because we cannot limit memory of the subprocesses directly AND there seem to be bugs in the Python 3.4 impl,
    ## could replace the subprocess call by a call to an external which would run under given memory/time limits
    ## and save output to given files.
    ## OR: update from Python 3.4.2?
    ## NAming output / model files: sort input filenames, replace spaces/punctuation
    ### USE smth like
    ## keepcharacters = (' ','.','_')
    ## "".join(c for c in filename if c.isalnum() or c in keepcharacters else 'I').strip()
    ## PARAMETERS
    ##   : solList: if not None, we are solving originally ( not checking )
    def solveInstance(self, s_Inst, slvBE, slvName, solList=None):
        resSlv = OrderedDict()
        bChkDZN = True if None!=solList and None!=self.params.args.checkDZN else False
        if bChkDZN:
            print( "_PARSING '", self.params.args.checkDZN, sep='', end="'... " )
            with open( self.params.args.checkDZN, 'r' ) as ro:
                mzn_exec.parseStdout( ro, resSlv, slvBE["Stdout_Keylines"], slvBE["Stdout_Keyvalues"], solList )
                print( ro.tell(), "bytes", end='' )
            if None!=self.params.args.checkStderr:
                print( "  and '", self.params.args.checkStderr, sep='', end="'... " )
                with open( self.params.args.checkStderr, 'r' ) as re:
                    mzn_exec.parseStderr( re, resSlv, slvBE["Stderr_Keylines"], slvBE["Stderr_Keyvalues"] )
                    print( re.tell(), "bytes", end='' )
        else:      #### Solving oneself
            print( slvName, "... ", sep='', end='', flush=True )
            s_Call = slvBE["EXE"]["s_SolverCall"][0] % s_Inst \
              + ' ' + slvBE["EXE"]["s_ExtraCmdline"][0]
            if self.params.args.addOption is not None:
                for sOpt in self.params.args.addOption:
                    s_Call += ' ' + sOpt
            if solList is not None:
                if self.params.args.addSolverOption is not None:
                    for sOpt in self.params.args.addSolverOption:
                        s_Call += ' ' + sOpt
            else:
                if self.params.args.addCheckerOption is not None:
                    for sOpt in self.params.args.addCheckerOption:
                        s_Call += ' ' + sOpt
            s_InstMerged = s_Inst.strip()
            ## s_InstMerged = regex.sub( r"[.\\/:~]", "", s_InstMerged );
            ## s_InstMerged = regex.sub( r"[ ]", "-", s_InstMerged );
            keepcharacters = ('-','_')
            s_InstMerged = "".join(c if c.isalnum() or c in keepcharacters else 'I' for c in s_InstMerged).strip()
            
            if solList is not None and self.params.args.useJoinedName is not None:
                for sUseJN in self.params.args.useJoinedName:
                    s_UsingOpt = sUseJN % s_InstMerged
                    s_Call += ' ' + s_UsingOpt
            if solList is not None:      ## solving the original instance
                sFlnStdout = sFlnStdoutBase.format( s_InstMerged )
                sFlnStderr = sFlnStderrBase.format( s_InstMerged )
                if self.params.args.debug is not None and ( self.params.args.debug & 1 ):
                    print( "  CALL: \"", s_Call, "\"", sep='', flush=True )
            else:
                sFlnStdout = 'last_stdout' + slvName + '.txt'
                sFlnStderr = 'last_stderr' + slvName + '.txt'
                if self.params.args.debug is not None and ( self.params.args.debug & 2 ):
                    print( "  CALL: \"", s_Call, "\"", sep='', flush=True )
            resSlv["Solver_Call"] = s_Call
            resSlv["DateTime_Start"] = datetime.datetime.now().__str__()
            if 1==self.params.args.psutils:
                completed, tmAll = \
                    mzn_exec.runCmd(
                      s_Call,
                      slvBE["EXE"]["b_ThruShell"][0],
                      slvBE["EXE"]["n_TimeoutRealHard"][0],
                      slvBE["EXE"]["n_VMEMLIMIT_SoftHard"]
                    )
                with utils.openFile_autoDir( sFlnStdout, "w" ) as tf:
                    tf.write( completed.stdout )
                with utils.openFile_autoDir( sFlnStderr, "w" ) as tf:
                    tf.write( completed.stderr )
                print( "STDOUT/ERR: ", len(completed.stdout), '/',
                      len(completed.stderr), " bytes", sep='', end=', ' )
                mzn_exec.parseStderr( io.StringIO( completed.stderr ), resSlv, slvBE["Stderr_Keylines"], slvBE["Stderr_Keyvalues"] )
                mzn_exec.parseStdout( io.StringIO( completed.stdout ), resSlv, slvBE["Stdout_Keylines"], slvBE["Stdout_Keyvalues"], solList )
                ## Adding the outputs to the log
                ## resSlv["StdErr"] = completed.stderr
                ## resSlv["StdOut"] = completed.stdout
            else:          ## use the 'system' call
                with utils.openFile_autoDir( sFlnStdout, "w" ) as tf:
                    tf.write( "% EMPTY\n" )
                with utils.openFile_autoDir( sFlnStderr, "w" ) as tf:
                    tf.write( "% EMPTY" )
                tmAll = mzn_exec.runCmdCmdline(
                      s_Call,
                      sFlnStdout, sFlnStderr,
                      self.params.cfg["COMMON_OPTIONS"]["runCommand"],
                      slvBE["EXE"]["n_TimeoutRealHard"][0],
                      (self.params.args.verbose) if solList is not None else (self.params.args.vc),
                      slvBE["EXE"]["n_VMEMLIMIT_SoftHard"]
                )
                with open( sFlnStderr, "r" ) as rf:
                    mzn_exec.parseStderr( rf, resSlv, slvBE["Stderr_Keylines"], slvBE["Stderr_Keyvalues"] )
                with open( sFlnStdout, "r" ) as rf:
                    mzn_exec.parseStdout( rf, resSlv, slvBE["Stdout_Keylines"], slvBE["Stdout_Keyvalues"], solList )

            print( " t: {:.3f}".format( tmAll ), end=' s, ' )
            resSlv["DateTime_Finish"] = datetime.datetime.now().__str__()
            resSlv["TimeReal_All"] = tmAll
            resSlv["TimeReal_LastStatus"] = 0
            resSlv["Hostname"] = platform.uname()[1]
        ### For all cases, some postprocessing #############################
        if "Sol_Status" not in resSlv:
            resSlv["Sol_Status"] = [-50, "   !!!!!  STATUS TOTALLY UNKNOWN - NO STATUS LINE PARSED."]
        if "Time_Flt" not in resSlv or utils.try_float( resSlv.get( "Time_Flt" ) ) is None:
            resSlv["NOFZN"] = ["      !!!!! No flattening finish time registered or successfully parsed"]
        dTmLast = utils.try_float( resSlv.get( "RealTime_Solns2Out" ) )
        if None!=dTmLast:
            resSlv["TimeReal_LastStatus"] = dTmLast
            resSlv.pop( "RealTime_Solns2Out" )
        ## if "SolutionLast" in resSlv:
        ##     print( "   SOLUTION_LAST:\n", resSlv["SolutionLast"], sep='' )
        if None!=solList:
            print( "  Nsol:", len(solList), end=',' )
        print( "   STATUS:", resSlv["Sol_Status"] )
        return resSlv
      
    ## Append the unchecked solution of the instance to a temp. file
    def saveSolution(self, wf):
        json_log.writeLogChunk( wf, json.dumps( self.result, indent=json_config.n_JSON_Indent ) )
    
    ## Return the necessity to check solutions.
    ## Can be false if we only want to compare different solvers.
    def ifShouldCheck(self):
        return \
          0<self.params.cfg["COMMON_OPTIONS"]["SOLUTION_CHECKING"]["n_FailedSaveMax"][0] and \
          0!=self.params.cfg["COMMON_OPTIONS"]["SOLUTION_CHECKING"]["n_CheckedMax"][0] and \
          0<len( self.params.chkBEs ) and \
          self.ifCheckableStatus( self.result["__SOLVE__"]["Sol_Status"][0] )
    
    ##
    def ifCheckableStatus( self, status ):
        return status>0
    
    ## Check selected solutions of the instance
    def checkOriginal( self, s_Inst ):
        nCM = self.params.cfg["COMMON_OPTIONS"]["SOLUTION_CHECKING"]["n_CheckedMax"][0]
        nFSM = self.params.cfg["COMMON_OPTIONS"]["SOLUTION_CHECKING"]["n_FailedSaveMax"][0]
        assert 0!=nCM
        assert 0<len(self.solList)
        rng = None
        if 0<nCM:
            rng = range( 0, min( nCM, len(self.solList ) ) )
        else:
            rng = range( -1, max( nCM-1, -len(self.solList)-1 ), -1 )
        self.result["SOLUTION_CHECKS_DONE"] = 0
        self.result["SOLUTION_CHECKS_FAILED"] = 0
        fFailed = 0
        ## Iterate over the solution range
        for iSol in rng:
            ## Try modify
            # self.solList[ iSol ] = self.solList[iSol][:20] + '1' + self.solList[iSol][21:]
            print ( "CHK SOL ", iSol if iSol<0 else (iSol+1), '/', len(rng), sep='', end='... ' )
            with utils.openFile_autoDir( sFlnSolLastDzn, "w" ) as wf:          ## Save the selected solution
                wf.write( self.solList[ iSol ] )
            s_IC = s_Inst + ' ' + sFlnSolLastDzn
            bCheckOK = True
            chkFlSt = []
            # self.result["__CHECKS__"] = []
            for iChk in range ( len ( self.params.chkBEs ) ):
                chkBE = self.params.chkBEs[ iChk ]
                chkRes = self.solveInstance( s_IC, chkBE, '__Checker_'+str(iChk+1) )
                # self.result["__CHECKS__"].append( chkRes )
                if ( ## -51==chkRes["Sol_Status"][0] or               ## NOFZN? No, flattener should report INFEAS.
                    ( 0>chkRes["Sol_Status"][0] and -3<=chkRes["Sol_Status"][0] ) ):   ## INFEAS
                    bCheckOK = False
                    chkFlSt = chkRes["Sol_Status"]
            self.result["SOLUTION_CHECKS_DONE"] += 1
            if not bCheckOK:
                fFailed = 1
                self.result["SOLUTION_CHECKS_FAILED"] += 1
                self.result["SOLUTION_FAILED_LAST"] = self.solList[ iSol ]
                # self.result["SOLUTION_FAILED_LAST__CHKSTATUS"] = chkRes["Sol_Status"]
                if self.fileFail is None:
                    self.fileFail = utils.openFile_autoDir( self.fileFailName, "w" )
                self.saveSolution( self.fileFail )
                if nFSM<=self.result["SOLUTION_CHECKS_FAILED"]:
                    print ( self.result["SOLUTION_CHECKS_FAILED"], "failed solution(s) saved, go on" )
                    break
              
        self.nCheckedInstances += 1
        self.nChecksFailed += fFailed
        print( "   CHECK FAILS on this instance: ", self.result["SOLUTION_CHECKS_FAILED"],
                ",  total check-failed instances: ", self.nChecksFailed,
                " from ", self.nCheckedInstances, sep='' )
            
    def __init__(self):
        ## Default params
        self.params = MZT_Param()

##############################################################################################
######################### MZNTest public #############################
##############################################################################################
    def run(self):
        self.obtainParams()
        self.compileExplicitModelLists()
        self.compileResultLogs()
        if not self.bCmpOnly:
            self.runTheInstances()
        else:
            self.compareLogs()
        self.summarize()
        
#  def main
mznTst = MznTest()
mznTst.run()




from pathlib import Path
from minizinc import default_driver, Driver
import subprocess
import json


def test_run_fzn():
    here = Path(__file__).resolve().parent
    assert isinstance(default_driver, Driver)
    model_file = here / "test_run_fzn.fzn"
    p = subprocess.run(
        [default_driver._executable, model_file, "--json-stream"],
        stdin=None,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    assert p.returncode == 0
    messages = [json.loads(l) for l in p.stdout.splitlines(False)]
    assert {
        "type": "solution",
        "output": {"dzn": "x = 1;\n"},
        "sections": ["dzn"],
    } in messages



from pathlib import Path
from minizinc import default_driver, Driver
import subprocess
import json


def test_model_interface_only_1():
    assert isinstance(default_driver, Driver)
    p = subprocess.Popen(
        [default_driver._executable, "-", "--model-interface-only",],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout, stderr = p.communicate(
        input=b'int: n; var 1..3: x; solve minimize x; output ["foo"];'
    )
    assert p.returncode == 0
    interface = json.loads(stdout)
    assert interface == {
        "type": "interface",
        "input": {"n": {"type": "int"}},
        "output": {"x": {"type": "int"}},
        "method": "min",
        "has_output_item": True,
        "included_files": [],
        "globals": [],
    }


def test_model_interface_only_2():
    assert isinstance(default_driver, Driver)
    p = subprocess.Popen(
        [default_driver._executable, "-", "--model-interface-only"],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout, stderr = p.communicate(input=b"int: n; var 1..3: x; solve maximize x;")
    assert p.returncode == 0
    interface = json.loads(stdout)
    assert interface == {
        "type": "interface",
        "input": {"n": {"type": "int"}},
        "output": {"x": {"type": "int"}},
        "method": "max",
        "has_output_item": False,
        "included_files": [],
        "globals": [],
    }


def test_model_interface_only_3():
    assert isinstance(default_driver, Driver)
    p = subprocess.Popen(
        [default_driver._executable, "-", "--model-interface-only"],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout, stderr = p.communicate(
        input=b'float: n; var set of 1..3: x; output ["foo"];'
    )
    assert p.returncode == 0
    interface = json.loads(stdout)
    assert interface == {
        "type": "interface",
        "input": {"n": {"type": "float"}},
        "output": {"x": {"type": "int", "set": True}},
        "method": "sat",
        "has_output_item": True,
        "included_files": [],
        "globals": [],
    }



from minizinc import default_driver, Driver
import json
import os


def test_tags_present():
    assert isinstance(default_driver, Driver)
    tags = ["mzn-fzn", "mzn-nl"]
    extra = os.environ.get("MZN_TEST_TAGS_PRESENT")
    if extra is not None:
        tags.extend(i.strip() for i in extra.split(","))
    for tag in tags:
        p = default_driver._run(["--solver-json", tag])
        solver = json.loads(p.stdout)
        assert "id" in solver



from . import yaml
from . import spec
from . import helpers

from flask import Flask, request, jsonify
from glob import glob
from datetime import timedelta

import re

app = Flask(__name__)


@app.route("/")
def root():
    return app.send_static_file("index.html")


@app.route("/files.json")
def files():
    files = [x[5:].replace("\\", "/") for x in glob("spec/**/*.mzn", recursive=True)]
    return jsonify({"files": files})


@app.route("/file.json")
def file():
    path = "./spec/" + request.args["f"]
    with open(path, encoding="utf-8") as file:
        contents = file.read()
        yaml_comment = re.match(r"\/\*\*\*\n(.*?)\n\*\*\*\/", contents, flags=re.S)
        if yaml_comment is None:
            return jsonify({"cases": []})

        tests = [x for x in yaml.load_all(yaml_comment.group(1))]

        cases = [
            {
                k: v
                for k, v in test.__dict__.items()
                if v is not yaml.Undefined
                and k in ["name", "solvers", "check_against", "markers", "type",]
            }
            for test in tests
        ]

        for case, test in zip(cases, tests):
            case["extra_files"] = "\n".join(test.extra_files)
            case["expected"] = (
                [{"value": yaml.dump(expected)} for expected in test.expected]
                if isinstance(test.expected, list)
                else [{"value": yaml.dump(test.expected)}]
            )

            case["options"] = [
                {"key": k, "value": (v if isinstance(v, str) else "")}
                for k, v in test.options.items()
                if k not in ["all_solutions", "timeout"]
            ]
            case["all_solutions"] = (
                "all_solutions" in test.options and test.options["all_solutions"]
            )
            case["timeout"] = (
                test.options["timeout"].total_seconds()
                if "timeout" in test.options
                else 0
            )
        return jsonify({"cases": cases})


def load_spec(data):
    items = {
        k: v
        for k, v in data.items()
        if k in ["name", "solvers", "check_against", "markers", "type"]
    }
    items["extra_files"] = data["extra_files"].splitlines(keepends=False)
    items["expected"] = [yaml.load(x["value"]) for x in data["expected"]]
    items["options"] = {
        x["key"]: (x["value"] if len(x["value"]) > 0 else True) for x in data["options"]
    }
    items["options"]["all_solutions"] = data["all_solutions"]
    if not data["timeout"] == "" and float(data["timeout"]) > 0:
        items["options"]["timeout"] = timedelta(seconds=float(data["timeout"]))
    print(items)
    return spec.Test(**items)


@app.route("/generate.json", methods=["POST"])
def generate():
    path = "./spec/" + request.args["f"]
    variables = request.args["vars"].splitlines(keepends=False)
    mode = request.args["mode"]
    data = request.json

    def filter_args(solution, mode, variables):
        if mode == "exclude":
            for x in variables:
                delattr(solution, x)
        else:
            all_vars = [x for x in solution.__dict__.keys()]
            for x in all_vars:
                if x not in variables:
                    delattr(solution, x)
        return solution

    test = load_spec(data)

    generated = []
    for solver in test.solvers:
        model, result, required, obtained = test.run(path, solver)
        if not test.passed(result):
            if isinstance(obtained, spec.Result):
                if isinstance(obtained.solution, list):
                    obtained.solution = [
                        filter_args(s, mode, variables) for s in obtained.solution
                    ]
                else:
                    filter_args(obtained.solution, mode, variables)
            test.expected.append(obtained)
            generated.append({"value": yaml.dump(obtained)})

    return jsonify({"obtained": generated})


@app.route("/save.json", methods=["POST"])
def save():
    path = "./spec/" + request.args["f"]
    data = request.json
    dumped = yaml.dump_all(load_spec(x) for x in data)

    with open(path, encoding="utf-8", mode="r+") as file:
        contents = file.read()
        yaml_comment = re.match(
            r"^(.*\/\*\*\*\n)(.*?)(\n\*\*\*\/.*)$", contents, flags=re.S
        )
        output = ""
        if yaml_comment is None:
            output = "/***\n" + dumped + "***/\n\n" + contents
        else:
            output = yaml_comment.group(1) + dumped + yaml_comment.group(3)

        file.seek(0)
        file.truncate()
        file.write(output)

    return jsonify({"status": "success"})


if __name__ == "__main__":
    app.run(debug=True)



from enum import Enum
from . import yaml
import minizinc as mzn
from minizinc.helpers import check_result
import pathlib
import re
from json import load, dumps
import jsonschema
import warnings


@yaml.mapping("!Test")
class Test:
    """
    Represents a test case

    Encoded using `!Test` in YAML.
    """

    def __init__(self, **kwargs):
        self.name = yaml.Undefined
        self.solvers = ["gecode", "cbc", "chuffed", "highs"]
        self.check_against = []
        self.expected = []
        self.options = {}
        self.extra_files = []
        self.markers = []
        self.type = "solve"

        for key, value in kwargs.items():
            setattr(self, key, value)

    def run(self, mzn_file, solver, default_options={}):
        """
        Runs this test case given an mzn file path, a solver name and some default options.

        Any options specified in this test case directly will override those provided in
        default options.

        Returns a tuple containing:
        - the model produced
        - the result of running the solver
        - a list of expected results
        - the actual obtained result

        Pass the actual obtained result to passed() to determine if the test case passed.
        """
        options = {k: v for k, v in default_options.items()}
        options.update(self.options)
        file = pathlib.Path(mzn_file)
        extra_files = [file.parent.joinpath(other) for other in self.extra_files]

        try:
            model = mzn.Model([file] + extra_files)
            model.output_type = Solution
            if solver.endswith(".msc"):
                solver = mzn.Solver.load(file.parent.joinpath(solver))
            else:
                solver = mzn.Solver.lookup(solver)
            instance = mzn.Instance(solver, model)
            if self.type == "solve":
                result = instance.solve(**options)
                obtained = Result.from_mzn(result)
            elif self.type == "compile":
                with instance.flat(**options) as (fzn, ozn, stats):
                    if solver.inputType == "JSON":
                        obtained = FlatZincJSON.from_mzn(fzn, file.parent)
                        result = obtained
                    else:
                        obtained = FlatZinc.from_mzn(fzn, file.parent)
                        result = obtained
            elif self.type == "output-model":
                with instance.flat(**options) as (fzn, ozn, stats):
                    obtained = OutputModel.from_mzn(ozn, file.parent)
                    result = obtained
            else:
                raise NotImplementedError("Unknown test case type")
        except mzn.MiniZincError as error:
            result = error
            obtained = Error.from_mzn(result)

        required = self.expected if isinstance(self.expected, list) else [self.expected]

        return model, result, required, obtained

    def passed(self, actual):
        """
        Returns whether the given result satisfies this test case
        """
        if isinstance(self.expected, list):
            return any(exp.check(actual) for exp in self.expected)
        return self.expected.check(actual)


@yaml.mapping("!Result")
class Result:
    """
    Represents a result generated by calling `solve()` on an instance.

    Encoded using `!Result` in YAML.
    """

    def __init__(self, **kwargs):
        self.status = yaml.Undefined
        self.solution = yaml.Undefined

        for k, v in kwargs.items():
            setattr(self, k, v)

    def check(self, actual):
        """
        Returns whether this result and the given result match.
        """
        if not isinstance(actual, mzn.Result):
            return False

        if self.status is not yaml.Undefined and str(actual.status) != self.status:
            return False

        if self.solution is not yaml.Undefined and self.solution is not None:
            required = self.solution
            obtained = actual.solution

            if isinstance(required, list):
                return len(required) == len(obtained) and all(
                    x.is_satisfied(y) for x, y in zip(required, obtained)
                )
            else:
                return required.is_satisfied(obtained)

        return True

    @staticmethod
    def from_mzn(result):
        """
        Constructs a Result from a minizinc Result type produced by running a solver
        as opposed to deserializing from YAML.
        """
        instance = Result()
        instance.status = str(result.status)
        instance.solution = result.solution
        return instance


@yaml.sequence("!SolutionSet")
class SolutionSet:
    """
    Represents a set of expected solutions that passes when
    every obtained solution matches an entry in this set.

    Represented with `!SolutionSet` in YAML
    """

    def __init__(self, *args):
        self.items = args

    def __iter__(self):
        return iter(self.items)

    def __len__(self):
        return len(self.items)

    def is_satisfied(self, other):
        """
        Returns whether or not every solution in `other` matches one in this solution set
        """
        return len(self) == len(other) and all(
            any(item.is_satisfied(o) for item in self.items) for o in other
        )


@yaml.mapping("!Solution")
class Solution:
    """
    A solution which is satisfied when every member of the this solution matches a member in the obtained solution.

    That is, the obtained solution can contain extra items that are not compared.

    Represented by `!Solution` in YAML.
    """

    def __init__(self, **kwargs):
        self.inner = {k: v for k, v in kwargs.items()}

    def items(self):
        return self.inner.items()

    def as_dict(self):
        return self.inner

    def is_satisfied(self, other):
        """
        Returns whether or not this solution is satisfied by an actual solution
        """

        def in_other(k, v):
            try:
                return v == other.inner[k]
            except KeyError:
                return False

        return all(in_other(k, v) for k, v in self.inner.items())


@yaml.mapping("!Error")
class Error:
    """
    A MiniZinc error with a type name and message.

    Represented by `!Error` in YAML.
    """

    def __init__(self, **kwargs):
        self.type = yaml.Undefined
        self.message = yaml.Undefined
        self.regex = yaml.Undefined

        for key, value in kwargs.items():
            setattr(self, key, value)

    def check(self, actual):
        """
        Checks if this error matches an actual obtained result/error
        """
        if not isinstance(actual, mzn.MiniZincError):
            return False

        if self.type is not yaml.Undefined:
            kind = type(actual)
            classes = (kind,) + kind.__bases__
            if not any(self.type == c.__name__ for c in classes):
                return False

        if self.message is not yaml.Undefined:
            return self.message == str(actual)

        if self.regex is not yaml.Undefined:
            return re.match(self.regex, str(actual), flags=re.M | re.S) is not None

        return True

    @staticmethod
    def from_mzn(error):
        """
        Creates an Error from a MiniZinc error (as opposed to from deserializing YAML)
        """
        instance = Error()
        instance.type = type(error).__name__
        instance.message = str(error)
        return instance


class CachedResult:
    """
    An object used to store a model and result from solving an instance

    This is so that `CheckItem`s do not have to re-solve an instance when
    verifying against another solver.
    """

    def __init__(self):
        self.model = None
        self.result = None
        self.obtained = None

    def test(self, solver):
        """
        Checks that the result stored in this object, when given as data for the
        stored model with the given solver, is satisfied.
        """
        solver_instance = mzn.Solver.lookup(solver)

        solution = (
            [s.as_dict() for s in self.result.solution]
            if isinstance(self.result.solution, list)
            else [self.result.solution.as_dict()]
        )
        passed = check_result(
            model=self.model,
            result=mzn.Result(self.result.status, solution, {}),
            solver=solver_instance,
        )
        return passed


@yaml.scalar("!FlatZinc")
class FlatZinc:
    """
    A FlatZinc result, encoded by !FlatZinc in YAML.
    """

    def __init__(self, path):
        self.path = path
        self.base = None
        self.fzn = None

    def check(self, actual):
        if not isinstance(actual, FlatZinc):
            return False

        if self.fzn is None:
            with open(actual.base.joinpath(self.path), encoding="utf-8") as f:
                self.fzn = f.read()

        return self.normalized() == actual.normalized()

    def normalized(self):
        """
        Quick and dirty normalization of FlatZinc

        Used to compare two `.fzn` files a bit more robustly than naive string comparison
        """
        fzn = self.fzn
        var_nums = re.findall(r"X_INTRODUCED_(\d+)_", fzn)
        for new, old in enumerate(sorted(set(int(x) for x in var_nums))):
            # Compress the introduced variable numbers, removing any holes and starting from 0
            needle = "X_INTRODUCED_{}_".format(old)
            replacement = "X_INTRODUCED_{}_".format(new)
            fzn = fzn.replace(needle, replacement)
        # Remove empty lines, truncate multiple spaces
        lines = [
            re.sub(
                r"::[^;]*",
                ann_sort,
                re.sub(r"  +", " ", line),
                count=1,
            )
            for line in fzn.split("\n")
            if len(line.strip()) > 0 and not line.strip().startswith("%")
        ]
        return "\n".join(sorted(lines))  # Sort the FlatZinc lines

    def get_value(self):
        if self.fzn is None:
            return self.path
        return self.normalized()

    @staticmethod
    def from_mzn(fzn, base):
        """
        Creates a `FlatZinc` object from a `File` returned by `flat()` in the minizinc interface.

        Also takes the base path the mzn file was from, so that when loading the expected fzn file
        it can be done relative to the mzn path.
        """
        with open(fzn.name, encoding="utf-8") as file:
            instance = FlatZinc(None)
            instance.fzn = file.read()
            instance.base = base
            return instance


def ann_sort(match):
    s = match.group(0)
    anns = sorted([ann for ann in s.split("::") if ann != ""])
    return "::" + "::".join(anns)


@yaml.scalar("!FlatZincJSON")
class FlatZincJSON:
    """
    A FlatZinc JSON result, encoded by !FlatZincJSON in YAML.
    """

    def __init__(self, path):
        self.path = path
        self.base = None
        self.fzn = None

    def check(self, actual):
        if not isinstance(actual, FlatZincJSON):
            return False

        if self.fzn is None:
            with open(actual.base.joinpath(self.path), encoding="utf-8") as f:
                self.fzn = load(f)

        obtained = actual.normalized()

        schema_path = (
            pathlib.Path(__file__).parent.parent.parent / "docs/en/fznjson.json"
        )
        if schema_path.exists():
            with open(schema_path, encoding="utf-8") as fp:
                schema = load(fp)
            jsonschema.validate(instance=obtained, schema=schema)
        else:
            warnings.warn(
                "Unable to find FlatZinc JSON schema. Cannot verify validity."
            )

        return self.normalized() == obtained

    def normalized(self):
        var_nums = set()
        for v in list(self.fzn["variables"].keys()) + list(self.fzn["arrays"].keys()):
            m = re.match(r"X_INTRODUCED_(\d+)_", v)
            if m is not None:
                var_nums.add(int(m.group(1)))
        var_map = {
            f"X_INTRODUCED_{src}_": f"X_INTRODUCED_{dst}_"
            for dst, src in enumerate(sorted(var_nums))
        }

        def get_ident(obj):
            if isinstance(obj, dict):
                return obj.get("id")
            return obj

        def walk(obj, key=None):
            if isinstance(obj, dict):
                return {walk(k): walk(v, k) for k, v in obj.items()}
            if isinstance(obj, list):
                items = [walk(v) for v in obj]
                if key == "ann":
                    return sorted(items, key=get_ident)
                return items
            if isinstance(obj, str):
                return var_map.get(obj, obj)
            return obj

        return walk(self.fzn)

    def get_value(self):
        if self.fzn is None:
            return self.path
        return dumps(self.normalized())

    @staticmethod
    def from_mzn(fzn, base):
        """
        Creates a `FlatZinc` object from a `File` returned by `flat()` in the minizinc interface.

        Also takes the base path the mzn file was from, so that when loading the expected fzn file
        it can be done relative to the mzn path.
        """
        with open(fzn.name, encoding="utf-8") as file:
            instance = FlatZincJSON(None)
            instance.fzn = load(file)
            instance.base = base
            return instance


def ann_sort(match):
    s = match.group(0)
    anns = sorted([ann for ann in s.split("::") if ann != ""])
    return "::" + "::".join(anns)


@yaml.scalar("!OutputModel")
class OutputModel:
    """
    An OZN result, encoded by !OutputModel in YAML.
    """

    def __init__(self, path):
        self.path = path
        self.base = None
        self.ozn = None

    def check(self, actual):
        if not isinstance(actual, OutputModel):
            return False

        if self.ozn is None:
            with open(actual.base.joinpath(self.path), encoding="utf-8") as f:
                self.ozn = f.read()

        return self.ozn == actual.ozn

    def get_value(self):
        if self.ozn is None:
            return self.path
        return self.ozn

    @staticmethod
    def from_mzn(ozn, base):
        """
        Creates a `OutputModel` object from a `File` returned by `flat()` in the minizinc interface.

        Also takes the base path the mzn file was from, so that when loading the expected ozn file
        it can be done relative to the mzn path.
        """
        with open(ozn.name, encoding="utf-8") as file:
            instance = OutputModel(None)
            instance.ozn = file.read()
            instance.base = base
            return instance


@yaml.mapping("!Suite")
class Suite:
    """
    Represents a test suite configuration

    Encoded using `!Suite` in YAML.
    """

    def __init__(self, **kwargs):
        self.solvers = None
        self.options = {}
        self.strict = True
        self.regex = r"."
        self.includes = ["*"]
        self.excludes = []

        for key, value in kwargs.items():
            setattr(self, key, value)



from pathlib import Path
import subprocess
import json


def test_mip_stats():
    from minizinc import default_driver, Driver

    assert isinstance(default_driver, Driver)
    model_file = Path(__file__).with_suffix(".mzn").resolve()

    p = subprocess.run(
        [
            default_driver._executable,
            model_file,
            "--json-stream",
            "--solver",
            "highs",
        ],
        stdin=None,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    actual = [json.loads(l) for l in p.stdout.splitlines(False)]
    if any(
        m["type"] == "error"
        and (m["what"] == "plugin loading error" or m["what"] == "configuration error")
        for m in actual
    ):
        import pytest

        pytest.skip("HiGHS not available")
        return
    assert p.returncode == 0
    stats = [
        s for m in actual if m["type"] == "statistics" for s in m["statistics"].keys()
    ]
    assert "objective" not in stats



from pathlib import Path
import subprocess
import json
import sys
import os
from tempfile import NamedTemporaryFile
from contextlib import contextmanager


@contextmanager
def named_temp_file(*args, **kwargs):
    # Workaround for temp files on Windows
    temp = NamedTemporaryFile(delete=False, *args, **kwargs)
    try:
        yield temp
    finally:
        temp.close()
        os.unlink(temp.name)


def test_json_stream():
    from minizinc import default_driver, Driver
    expected = [
        {"type": "trace", "section": "default", "message": "foo\n"},
        {"type": "comment", "comment": "% This is a comment\n"},
        {"type": "statistics", "statistics": {"foo": 123, "bar": "string"}},
        {
            "type": "solution",
            "output": {"dzn": "x = 1;\ny = 1;\n"},
            "sections": ["dzn"],
        },
        {
            "type": "solution",
            "output": {"dzn": "x = 1;\ny = 2;\n"},
            "sections": ["dzn"],
        },
        {
            "type": "solution",
            "output": {"dzn": "x = 1;\ny = 3;\n"},
            "sections": ["dzn"],
        },
        {
            "type": "solution",
            "output": {"dzn": "x = 2;\ny = 3;\n"},
            "sections": ["dzn"],
        },
        {
            "type": "solution",
            "output": {"dzn": "x = 3;\ny = 3;\n"},
            "sections": ["dzn"],
        },
        {"type": "status", "status": "OPTIMAL_SOLUTION"},
    ]

    here = Path(__file__).resolve().parent
    assert isinstance(default_driver, Driver)
    model_file = here / "test_json_stream.mzn"
    with named_temp_file(suffix=".msc", mode="w", encoding="utf-8") as fp:
        json.dump(
            {
                "name": "Test solver",
                "version": "1.0",
                "id": "org.minizinc.test_solver",
                "executable": [
                    Path(sys.executable).resolve().as_posix(),
                    Path(__file__).resolve().as_posix(),
                ],
            },
            fp,
        )
        fp.close()
        p = subprocess.run(
            [
                default_driver._executable,
                model_file,
                "--json-stream",
                "--solver",
                fp.name,
            ],
            stdin=None,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        assert p.returncode == 0
        actual = [json.loads(l) for l in p.stdout.splitlines(False)]
        assert expected == actual


if __name__ == "__main__":
    # Dummy solver output
    print("% This is a comment")
    print("%%%mzn-stat: foo=123")
    print('%%%mzn-stat: bar="string"')
    print("%%%mzn-stat-end")
    obj = 0
    for x in range(1, 4):
        for y in range(1, 4):
            if x + y > obj:
                obj = x + y
                print(f"x = {x};")
                print(f"y = {y};")
                print("----------")
    print("==========")



from pathlib import Path
from minizinc import default_driver, Driver
import subprocess
import json


def test_multiple_type_errors():
    here = Path(__file__).resolve().parent
    assert isinstance(default_driver, Driver)
    model_file = here / "test_multiple_type_errors.mzn"
    p = subprocess.run(
        [default_driver._executable, model_file, "--json-stream"],
        stdin=None,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    assert p.returncode == 1
    messages = [json.loads(l) for l in p.stdout.splitlines(False)]
    errors = [m for m in messages if m["type"] == "error"]
    assert len(errors) == 4
    assert all(e["what"] == "type error" for e in errors)



import timeit, re, sys, os
import utils, json_config

## TODO Keyline/value dictionaries: entries == json_config.s_CommentKey are ignored. Make it a parameter

class Output:
    def __init__( self ):
        stdout = ""
        stderr = ""

def on_terminate(proc):
    print("process {} terminated with exit code {}".format(proc, proc.returncode))

##############################################################################################
###################### MZN instance execution + result parsing low-level #####################
##############################################################################################
## runCmdCmdline using system(). Actually used for checking as well.
## run the specified shell command string and return the time.
## can add ulimit etc.
## s1, s2: filenames for strout, stderr
## dictCmd: commands for Win and non-Win
## meml: list of 2 values, soft & hard limits as N bytes
def runCmdCmdline( s_Cmd, s1, s2, dictCmd, timeo, bVerbose=False, meml=None ):
    tm = timeit.default_timer()
    setCmd = dictCmd["windows"] if "win" in sys.platform and "cygwin" not in sys.platform else dictCmd["non-windows"]
    sCmd = setCmd["runVerbose" if bVerbose else "runSilent"].format(meml[1], timeo, s_Cmd, s1, s2)
    print( "\n  RUNNING:", sCmd )
    os.system(sCmd)
    tm = timeit.default_timer() - tm
    return tm


## runCmd using psutils. Actually should be used for checking as well.
## run the specified shell command string and return the popen result.
## TODO catch intermediate solutions if needed
## meml: list of 2 values, soft & hard limits as N bytes
def runCmd( s_Cmd, b_Shell=False, timeo=None, meml=None ):
    import psutil, shlex, subprocess, resource
    if b_Shell:
        l_Cmd = s_Cmd
    else:
        l_Cmd = shlex.split( s_Cmd )
    tm = timeit.default_timer()
    ################# In the following, the subprocess.RUN method fails to kill shell calls on Linux
    #try:
    #    completed = subprocess.run( l_Cmd, shell=b_Shell,
    #      universal_newlines=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, timeout=timeo )
    #except subprocess.TimeoutExpired as te:
    #    completed = te
    ################# Using psutils
    proc = psutil.Popen(l_Cmd, shell=b_Shell,
          universal_newlines=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if None!=meml:
        if hasattr( psutil, 'RLIMIT_AS' ):  ## TODO move into preexec_fn for parallel tests?
            proc.rlimit( resource.RLIMIT_AS, ( meml[0]*1000, meml[1]*1000 ) )
        else:
            print( "  ... but the OS doesn't support RLIMIT_AS." )
    completed = Output()
    try:
        completed.stdout, completed.stderr = proc.communicate(timeout=timeo)
    except subprocess.TimeoutExpired:
        print ( " soft_kill. ", end='' )
        procs = psutil.Process().children(recursive=True)
        for p in procs:
            p.terminate()
        try:
            completed.stdout, completed.stderr = proc.communicate(timeout=1)
        except subprocess.TimeoutExpired as te:
            print ( " hard_kill. ", end='' )
            procs = psutil.Process().children(recursive=True)
            for p in procs:
                p.kill()
            completed.stdout, completed.stderr = proc.communicate()
    ### Clean up: (does psutil.Process.communicate() wait for all descendants?)  -----------------------------------
    procs = psutil.Process().children(recursive=True)
    for p in procs:
        p.kill()
    ### OR: even Queue?
#    with psutil.Popen(["ifconfig"], stdout=subprocess.PIPE) as proc:
#        log.write(proc.stdout.read())
#
#    procs = psutil.Process().children()
#    for p in procs:
#        p.terminate()
#    gone, still_alive = psutil.wait_procs(procs, timeout=3, callback=on_terminate)
#    for p in still_alive:
#        p.kill()
    tm = timeit.default_timer() - tm
    return completed, tm
    
def parseStderr( f, result, mapKL, mapKV ):
#        result["ProbSense"] = None
#        result["TimeFlt"] = None
    for line in f:
        line = line.strip()
        checkKeylines( line, mapKL, result )
        checkKeyvalues( line, mapKV, result )
        
## Puts feasible solutions into solList if it's not None
def parseStdout( f, result, mapKL, mapKV, solList ):
    l_SolLast = ""
    n_SolStatus = 0
    result["Number_Solutions"] = 0
    for line in f:
        line = line.rstrip()             ## To remove \n and spaces on the right
        res00 = {}               ## A temporary to see if we get a feasible solution separator
        checkKeylines( line, mapKL, res00 )
        utils.mergeDict( result, res00 )
        checkKeyvalues( line, mapKV, result )
        ## See if it's a solution status
        if "Sol_Status" in res00:
            if 1==res00[ "Sol_Status" ][0]:
                result["Number_Solutions"] += 1
##                result["Solution_Last"] = l_SolLast     ## Or save it? Need from a file then but might be great to have here
                if None!=solList:
                    solList.append( l_SolLast )
                l_SolLast = ""                         ## Clean up
        else:
            l_SolLast += line
            l_SolLast += '\n'
            
## Check if any keylines of the given 2-level dictionary equal the given line
## The 1st level gives variable name, 2nd level gives the line->value map
def checkKeylines( line, dict2, result ):
    assert dict==type(dict2)     ## What if we use OrderedDict???     TODO
    assert isinstance(result, dict)
    for key in dict2:
        if json_config.s_CommentKey!=key:
            val1 = dict2[key]
            assert dict==type(val1), "checkKeylines: key '%s': '%s'==type (%s)" % \
              ( key, type(val1), val1.__str__() )
            if line in val1:
                result[ key ] = [ val1[ line ], line ]
    
## Check if any search pattern of the given dict values is in the given line
## The key gives variable name, value gives the line->value mapping
def checkKeyvalues( line, dictVal, result ):
    assert dict==type(dictVal)     ## What if we use OrderedDict???     TODO
    assert isinstance(result, dict)
    for key in dictVal:
        if json_config.s_CommentKey!=key:
            paramArray = dictVal[ key ]
            assert 3<=len( paramArray ), \
              "Key '%s': param list should have >=3 elements, now: %s" % (key, paramArray.__str__())
            if None!=re.search( paramArray[0], line ):
                try:
                    lineSubst = re.sub( paramArray[1], ' ', line )
                except:
                    print("   WARNING: failed to substitute regex '", paramArray[1],
                          "' by ' ' in string '", line, "':  ", sys.exc_info()[0:2]
                          , sep='')
                else:
                    lSL = lineSubst.split()
                    if len( lSL )>paramArray[2]-1:
                        s_Val = lSL[ paramArray[2]-1 ]
 ##               d_Val = try_float( s_Val ) 
                        result[key] = s_Val             ### [ d_Val, s_Val ] Need here?
##                        print( "                checkKeyval: result[{}] = '{}'".format(key, s_Val) )
                    else:
                        print( "ERROR: Parsing output line ", lSL,
                              ": regex key '", paramArray[0],
                              "' found, but the split line too short for index ", paramArray[2],
                              sep='' )



import yaml

try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper  # type: ignore

import datetime


class Undefined:
    """
    Represents missing values in YAML (as opposed to null which is None in Python)
    """

    pass


def mapping(tag):
    """
    A decorator which allows for serializing/deserializing a class as a YAML mapping (dictionary).
    The class must be able to be constructed with `**kwargs` to set its data.

    :param tag: The tag to use for the type (e.g. `!MyTag`)
    :type tag: str
    """

    def decorator(obj_class):
        def construct(loader, node):
            try:
                attrs = loader.construct_mapping(node)
                instance = obj_class(**attrs)
                return instance
            except yaml.constructor.ConstructorError:
                return obj_class()

        def represent(dumper, data):
            items = data.items() if hasattr(data, "items") else data.__dict__.items()
            attrs = {key: value for key, value in items if value is not Undefined}
            return dumper.represent_mapping(tag, attrs)

        yaml.add_constructor(tag, construct, Loader=Loader)
        yaml.add_representer(obj_class, represent, Dumper=Dumper)
        return obj_class

    return decorator


def sequence(tag):
    """
    A decorator which allows for serializing/deserializing a class as a YAML sequence (list).
    The class must be able to be constructed with `*args` to set it the items, and also be iterable.

    :param tag: The tag to use for the type (e.g. `!MyTag`)
    :type tag: str
    """

    def decorator(obj_class):
        def construct(loader, node):
            try:
                args = loader.construct_sequence(node)
                return obj_class(*args)
            except yaml.constructor.ConstructorError:
                return obj_class()

        def represent(dumper, data):
            return dumper.represent_sequence(tag, iter(data))

        yaml.add_constructor(tag, construct, Loader=Loader)
        yaml.add_representer(obj_class, represent, Dumper=Dumper)
        return obj_class

    return decorator


def scalar(tag):
    """
    A decorator which allows for serializing/deserializing a class as a YAML scalar (string).
    The class must be able to be constructed with a string argument, and implement `get_value()`
    to return the string to be serialized.

    :param tag: The tag to use for the type (e.g. `!MyTag`)
    :type tag: str
    """

    def decorator(obj_class):
        def construct(loader, node):
            value = loader.construct_scalar(node)
            return obj_class(value)

        def represent(dumper, data):
            return dumper.represent_scalar(tag, str(data.get_value()))

        yaml.add_constructor(tag, construct, Loader=Loader)
        yaml.add_representer(obj_class, represent, Dumper=Dumper)
        return obj_class

    return decorator


def load(stream):
    """
    Helper function which loads YAML
    """
    return yaml.load(stream, Loader=Loader)


def load_all(stream):
    """
    Helper function which loads YAML as a list of documents
    """
    return yaml.load_all(stream, Loader=Loader)


def dump(data):
    """
    Helper function which serializes objects as YAML
    """
    return yaml.dump(data, Dumper=Dumper)


def dump_all(data):
    """
    Helper function which serializes a list of objects as YAML
    """
    return yaml.dump_all(data, Dumper=Dumper)


def range_representer(dumper, data):
    """
    A YAML `!Range l..u` tag
    """
    scalar = "{}..{}".format(data.start, data.stop - 1)
    return dumper.represent_scalar("!Range", scalar)


yaml.add_representer(range, range_representer, Dumper=Dumper)


@scalar("!Range")
class SetRange:
    def __init__(self, value):
        a, b = map(int, value.split(".."))
        self.value = range(a, b + 1)

    def __eq__(self, other):
        if isinstance(other, SetRange):
            other = other.value
        if isinstance(other, range):
            other = set(other)
        return set(self.value) == other

    def __repr__(self):
        return f"{self.value}"

    def get_value(self):
        return f"{self.value.start}..{self.value.stop - 1}"


def dt_representer(dumper, data):
    """
    A YAML `!Duration` tag
    """
    scalar = "{}ms".format(data.total_seconds() * 1000)
    return dumper.represent_scalar("!Duration", scalar)


yaml.add_representer(datetime.timedelta, dt_representer, Dumper=Dumper)


def dt_constructor(loader, node):
    """
    A YAML `!Duration` tag
    """
    value = loader.construct_scalar(node)
    if value.endswith("us"):
        return datetime.timedelta(microseconds=float(value[:-2]))
    if value.endswith("ms"):
        return datetime.timedelta(milliseconds=float(value[:-2]))
    if value.endswith("s"):
        return datetime.timedelta(seconds=float(value[:-1]))


yaml.add_constructor("!Duration", dt_constructor, Loader=Loader)


def list_representer(dumper, data):
    """
    Changes flow style of lists to be on a single line if only primitives are contained
    """
    flow_style = all(isinstance(i, (int, float, str, bool)) for i in data)
    return dumper.represent_sequence(
        "tag:yaml.org,2002:seq", data, flow_style=flow_style
    )


yaml.add_representer(list, list_representer)
